# coding: utf-8
"""
More operators for onnx reference implementation.
Experimentation with openmp, CUDA.
"""

__version__ = "0.2.2"
__author__ = "Xavier Dupré"


def has_cuda() -> bool:
    """
    Tells if cuda is available.
    """
    from ._config import HAS_CUDA

    return HAS_CUDA == 1


def cuda_version() -> str:
    """
    Tells which version of CUDA was used to build the CUDA extensions.
    """
    if not has_cuda():
        raise RuntimeError("CUDA extensions are not available.")
    from ._config import CUDA_VERSION

    return CUDA_VERSION


def cuda_version_int() -> tuple:
    """
    Tells which version of CUDA was used to build the CUDA extensions.
    It returns `(0, 0)` if CUDA is not present.
    """
    if not has_cuda():
        return (0, 0)
    from ._config import CUDA_VERSION

    if not isinstance(CUDA_VERSION, str):
        return tuple()

    spl = CUDA_VERSION.split(".")
    return tuple(map(int, spl))


def compiled_with_cuda() -> bool:
    """
    Checks it was compiled with CUDA.
    """
    try:
        from .validation.cuda import cuda_example_py

        return cuda_example_py is not None
    except ImportError:
        return False


def get_cxx_flags() -> str:
    """
    Returns `CXX_FLAGS`.
    """
    from ._config import CXX_FLAGS

    return CXX_FLAGS


def get_stdcpp() -> int:
    """
    Returns `CMAKE_CXX_STANDARD`.
    """
    from ._config import CMAKE_CXX_STANDARD

    return CMAKE_CXX_STANDARD



import os
import re
import unittest
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unittest.case import SkipTest
import numpy
import onnx.backend.base
import onnx.backend.test
import onnx.shape_inference
import onnx.version_converter
from onnx import FunctionProto, ModelProto, NodeProto
from onnx.backend.test import BackendTest
from onnx.backend.test.runner import Pattern, TestItem
from onnx.backend.test.loader import load_model_tests
from onnx.backend.base import Backend, Device, DeviceType
from .c_reference_evaluator import CReferenceEvaluator


class Runner:
    """
    Collects tests and run them as unit tests.

    :param backend: a subclass of :class:`onnx.backend.base.Backend`
    :param path_to_test: folder to look at
    :param kind: subfolder to test
    :param test_kwargs: additional test parameters
    """

    _add_model_test = BackendTest._add_model_test
    _add_test = BackendTest._add_test
    _load_proto = BackendTest._load_proto
    assert_similar_outputs = BackendTest.assert_similar_outputs

    def __init__(
        self,
        backend: type[Backend],
        path_to_test: Optional[str] = None,
        kind: Optional[Union[str, List[str]]] = None,
        test_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        self.backend = backend
        self._include_patterns: Set[Pattern[str]] = set()
        self._exclude_patterns: Set[Pattern[str]] = set()
        self._xfail_patterns: Set[Pattern[str]] = set()
        self._test_kwargs: dict = test_kwargs or {}

        if path_to_test is None or not os.path.exists(path_to_test):
            raise FileNotFoundError(f"Unable to find path {path_to_test!r}.")
        if isinstance(kind, str):
            kind = [kind]
        elif kind is None:
            kind = [
                k
                for k in os.listdir(path_to_test)
                if os.path.isdir(os.path.join(path_to_test, k))
            ]

        self._test_items: Dict[str, Dict[str, TestItem]] = {
            f"{k}Model": {} for k in kind
        }

        for k in kind:
            for ot in load_model_tests(path_to_test, kind=k):
                self._add_model_test(ot, k)

    def include(self, pattern: str) -> "Runner":
        self._include_patterns.add(re.compile(pattern))
        return self

    def exclude(self, pattern: str) -> "Runner":
        self._exclude_patterns.add(re.compile(pattern))
        return self

    def xfail(self, pattern: str) -> "Runner":
        self._xfail_patterns.add(re.compile(pattern))
        return self

    def _filtered_test_items(self) -> dict[str, dict[str, TestItem]]:
        filtered: dict[str, dict[str, TestItem]] = {}
        for category, items_map in self._test_items.items():
            filtered[category] = {}
            for name, item in items_map.items():
                if self._include_patterns and (
                    not any(include.search(name) for include in self._include_patterns)
                ):
                    item.func = unittest.skip("no matched include pattern")(item.func)
                for exclude in self._exclude_patterns:
                    if exclude.search(name):
                        item.func = unittest.skip(
                            f"matched exclude pattern '{exclude.pattern}'"
                        )(item.func)
                for xfail in self._xfail_patterns:
                    if xfail.search(name):
                        item.func = unittest.expectedFailure(item.func)
                filtered[category][name] = item
        return filtered

    def tests(self, name: str = "CustomTestCase") -> type[unittest.TestCase]:
        """
        Returns a subclass of `unittest.TestCase`.

        :param name: name of the subclass
        """
        tests = type("CustomTestCase", (unittest.TestCase,), {})
        for items_map in sorted(
            self._filtered_test_items().values(), key=lambda cl: cl.__class__.__name__
        ):
            for name, item in sorted(items_map.items()):
                setattr(tests, name, item.func)
        return tests

    def run(
        self, verbose: int = 0, exc_cls: Optional[type] = AssertionError
    ) -> Tuple[
        List[Tuple[str, Callable]],
        List[Tuple[str, Callable, Any]],
        List[Tuple[str, Callable, Exception]],
    ]:
        """
        Runs all tests.

        :param verbose: verbosity, use :epkg:`tqdm`
        :param exc_cls: exception to raise when a test fails, if None,
            no exception is raised
        :return: list of run tests, list of skipped tests, list of failed tests
        """
        tests = self.tests()
        methods = []
        for att in dir(tests):
            if att.startswith("test_"):
                test = getattr(tests, att)
                methods.append((att, test))
        if len(methods) == 0:
            msg = "\n".join(dir(tests))
            raise RuntimeError(f"No test was detected. Available tests are:\n{msg}")

        if verbose:
            from tqdm import tqdm

            loop = tqdm(methods)
        else:
            loop = methods

        ran = []
        skipped = []
        failed = []
        for i, (name, f) in enumerate(loop):
            if verbose:
                loop.set_description(f"{i+1}/{len(methods)}-{name}")
            try:
                f(tests)
            except SkipTest as es:
                skipped.append((name, f, es))
                continue
            except Exception as e:
                if exc_cls is not None:
                    raise exc_cls(f"Test {i}-{name!r} failed.") from e
                failed.append((name, f, e))
            ran.append((name, f))
        return ran, skipped, failed


class CReferenceEvaluatorBackendRep(onnx.backend.base.BackendRep):
    """
    See :class:`onnx_extended.reference.CReferenceEvaluator`
    for an example.

    :param session: any runtime with the same interface as
        :class:`onnx.reference.ReferenceEvaluator`
    """

    def __init__(self, session: CReferenceEvaluator):
        self._session = session

    def run(self, inputs: List[numpy.ndarray], **kwargs) -> List[numpy.ndarray]:
        if isinstance(inputs, numpy.ndarray):
            inputs = [inputs]
        if isinstance(inputs, list):
            if len(inputs) == len(self._session.input_names):
                feeds = dict(zip(self._session.input_names, inputs))
            else:
                feeds = {}
                pos_inputs = 0
                for inp, tshape in zip(
                    self._session.input_names, self._session.input_types
                ):
                    shape = tuple(d.dim_value for d in tshape.tensor_type.shape.dim)
                    if shape == inputs[pos_inputs].shape:
                        feeds[inp] = inputs[pos_inputs]
                        pos_inputs += 1
                        if pos_inputs >= len(inputs):
                            break
        elif isinstance(inputs, dict):
            feeds = inputs
        else:
            raise TypeError(f"Unexpected input type {type(inputs)!r}.")
        outs = self._session.run(None, feeds)
        return outs


class CReferenceEvaluatorBackend(onnx.backend.base.Backend):
    """
    See :class:`onnx_extended.reference.CReferenceEvaluator`
    for an example.
    """

    cls_inference = CReferenceEvaluator

    @classmethod
    def __class_getitem__(cls, cls_inference: type, name: Optional[str] = None) -> type:
        """
        Creates a new class inheriting from this one but with
        static attribute `cls_inference` equal to *cls_inference*.
        The goal is to make it easier to evaluate a runtime
        sharing the same API as the :class:`CReferenceEvaluator`
        on CPU.
        """
        if name is None:
            name = f"{cls.__name__}{cls_inference.__name__}"
        return type(name, (cls,), {"cls_inference": cls_inference})

    @classmethod
    def is_opset_supported(cls, model):
        """
        Tells which opsets are supported.
        """
        return True, ""

    @classmethod
    def supports_device(cls, device: str) -> bool:
        """
        Tells if a specific device is supported.
        """
        d = Device(device)
        return d.type == DeviceType.CPU

    @classmethod
    def create_inference_session(
        cls, model: Union[str, bytes, ModelProto, NodeProto, FunctionProto]
    ):
        """
        Creates an instance of the class running a model.
        """
        return cls.cls_inference(model)

    @classmethod
    def prepare(
        cls, model: Any, device: str = "CPU", **kwargs: Dict[str, Any]
    ) -> CReferenceEvaluatorBackendRep:
        if isinstance(model, cls.cls_inference):
            return CReferenceEvaluatorBackendRep(model)
        if isinstance(model, (str, bytes, ModelProto)):
            inf = cls.create_inference_session(model)
            return cls.prepare(inf, device, **kwargs)
        raise TypeError(f"Unexpected type {type(model)} for model.")

    @classmethod
    def run_model(
        cls,
        model,
        inputs: List[Any],
        device: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ):
        """
        Called if the onnx proto is a `ModelProto`.
        """
        rep = cls.prepare(model, device or "cpu", **kwargs)
        return rep.run(inputs, **kwargs)

    @classmethod
    def run_node(cls, node, inputs, device=None, outputs_info=None, **kwargs):
        """
        Called if the onnx proto is a `NodeProto`.
        """
        raise NotImplementedError("Unable to run the model node by node.")


def create_reference_backend(
    backend: Optional[type[Backend]] = None,
    path_to_test: Optional[str] = None,
    kind: Optional[str] = None,
) -> Runner:
    return Runner(
        backend or CReferenceEvaluatorBackend,
        path_to_test=path_to_test,
        kind=kind,
    )



import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Union
from onnx import (
    AttributeProto,
    FunctionProto,
    GraphProto,
    ModelProto,
    NodeProto,
    ValueInfoProto,
    TensorProto,
    TypeProto,
    shape_inference,
)
from onnx.helper import (
    make_attribute,
    make_graph,
    make_function,
    make_model,
    make_tensor_value_info,
    np_dtype_to_tensor_dtype,
    set_model_props,
)

logger = logging.getLogger("onnx-extended")


def _make_att_graph(name: str, new_body: GraphProto) -> AttributeProto:
    attr = AttributeProto()
    attr.name = name
    attr.g.CopyFrom(new_body)
    attr.type = AttributeProto.GRAPH
    return attr


def _make_node(
    op_type: str,
    inputs: List[str],
    outputs: List[str],
    name: Optional[str] = None,
    doc_string: Optional[str] = None,
    domain: str = "",
    attributes: Optional[Dict[str, Any]] = None,
) -> NodeProto:
    """
    Constructs a NodeProto.

    :param op_type: (string): The name of the operator to construct
    :param inputs: list of input names
    :param outputs: list of output names
    :param name: optional unique identifier for NodeProto
    :param doc_string: optional documentation
        string for NodeProto
    :param domain: optional domain for NodeProto.
        If it's None, we will just use default domain (which is empty)
    :param attributes: the attributes of the node. The acceptable values
        are documented in `make_attribute`.
    :return: node
    """
    node = NodeProto()
    node.op_type = op_type
    node.input.extend(inputs)
    node.output.extend(outputs)
    if name:
        node.name = name
    if doc_string:
        node.doc_string = doc_string
    if domain is not None:
        node.domain = domain
    if isinstance(attributes, dict):
        if len(attributes) > 0:
            node.attribute.extend(
                make_attribute(key, value) for key, value in sorted(attributes.items())
            )
    elif attributes:
        for att in attributes:
            node.attribute.extend([att])
    return node


def _apply_optimisation_on_graph(
    fct: Callable,
    onnx_model: Union[ModelProto, GraphProto, FunctionProto],
    recursive: bool = True,
    debug_info: Optional[List[str]] = None,
    **kwargs: Dict[str, Any],
) -> Union[ModelProto, GraphProto, FunctionProto]:
    """
    Applies an optimisation function *fct* on a graph
    and not on the model.

    :param fct: function to optimize like :func:`onnx_remove_node_unused`
    :param onnx_model: onnx model
    :param recursive: looks into subgraphs
    :param debug_info: debug information (private)
    :param kwargs: additional parameters
    :return: new onnx _model
    """
    if hasattr(onnx_model, "graph"):
        if debug_info is None:
            debug_info = []
        graph = fct(onnx_model.graph, debug_info=debug_info + ["GRAPH"], **kwargs)
        new_model = make_model(graph, functions=onnx_model.functions)
        new_model.ir_version = onnx_model.ir_version
        new_model.producer_name = onnx_model.producer_name
        new_model.producer_version = onnx_model.producer_version
        new_model.domain = onnx_model.domain
        new_model.model_version = onnx_model.model_version
        new_model.doc_string = onnx_model.doc_string
        if hasattr(onnx_model, "value_info"):
            graph.value_info.extend(onnx_model.value_info)
        while len(new_model.opset_import) > 0:
            new_model.opset_import.pop()
        for oimp in onnx_model.opset_import:
            op_set = new_model.opset_import.add()
            op_set.domain = oimp.domain
            op_set.version = oimp.version
        return new_model
    raise TypeError(
        f"This function only works on 'ModelProto' anod not not on {type(onnx_model)}."
    )


def _apply_remove_node_fct_node(
    fct: Callable, node: NodeProto, recursive: bool, debug_info: str
) -> NodeProto:
    """
    Applies an optimizing function on a subgraphs.

    :param node: onnx node
    :param recursive: does it in subgraphs as well
    :return: new node
    """
    if not hasattr(node, "attribute"):
        return node
    modified = 0
    new_atts = []
    for att in node.attribute:
        if att.name in ("body", "then_branch", "else_branch"):
            new_body = fct(
                att.g, recursive=recursive, debug_info=debug_info + [att.name]
            )
            new_atts.append(_make_att_graph(att.name, new_body))
            modified += 1
        else:
            new_atts.append(att)
    if modified > 0:
        new_node = _make_node(
            node.op_type, node.input, node.output, name=node.name, attributes=new_atts
        )
        return new_node
    return node


def _process_node(
    node: NodeProto,
    data: Dict,
    edges: Dict,
    paths: Dict,
    prefix: str = "",
    sep: str = ":X:",
    path: Optional[List[str]] = None,
):
    node_name = prefix + node.name
    data[node_name, 1] = node
    path = [] if path is None else path.copy()
    paths[node_name, 1] = path
    path = path.copy()
    path.append(node_name)
    for inp in node.input:
        data[inp, 0] = node
        edges[(inp, 0), (node_name, 1)] = node
        paths[inp, 0] = path
        if sep in node_name:
            # We need to link an input to the parent node
            # if the node is part of subgraph.
            # path_r = paths[inp, 0]
            if len(path) <= 1:
                raise RuntimeError(
                    f"Unexpected path {path!r}, this may happen "
                    f"if sep={sep!r} is already used in the original model."
                )
            edges[(inp, 0), (path[-2], 1)] = node

    for out in node.output:
        data[out, 0] = node
        paths[out, 0] = node_name
        edges[(node_name, 1), (out, 0)] = node
    if len(node.attribute) > 0:
        for att in node.attribute:
            if not hasattr(att, "g"):
                continue
            if not isinstance(att.g, GraphProto):
                continue
            for no in att.g.node:
                _process_node(no, data, edges, paths, prefix=node_name + sep, path=path)


def onnx_remove_node_unused(onnx_model, recursive=True, debug_info=None, **options):
    """
    Removes unused nodes of the graph. An unused node
    is not involved in the output computation.

    :param onnx_model: onnx model
    :param recursive: looks into subgraphs
    :param debug_info: debug information (private)
    :param options: unused
    :return: new onnx _model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).rsplit(".", maxsplit=1)[-1].strip("'>")]
    else:
        debug_info = debug_info + [
            str(type(onnx_model)).rsplit(".", maxsplit=1)[-1].strip("'>")
        ]

    if hasattr(onnx_model, "graph"):
        return _apply_optimisation_on_graph(
            onnx_remove_node_unused,
            onnx_model,
            recursive=recursive,
            debug_info=debug_info,
            **options,
        )

    graph = onnx_model
    logger.debug("onnx_remove_node_unused:begin with %d nodes.", len(graph.node))
    is_function = isinstance(graph, FunctionProto)
    data = {}
    valid = {}
    edges = {}
    paths = {}

    if not is_function:
        for init in graph.initializer:
            data[init.name, 0] = init

    for node in graph.node:
        _process_node(node, data, edges, paths)

    for out in graph.output:
        valid[out if is_function else out.name, 0] = True

    modif = 1
    while modif > 0:
        modif = 0
        for e1, e2 in edges:  # pylint: disable=E1141
            if valid.get(e2, False) and not valid.get(e1, False):
                valid[e1] = True
                modif += 1

    new_nodes = [n for n in graph.node if (n.name, 1) in valid]
    if not is_function:
        new_inits = [n for n in graph.initializer if (n.name, 0) in valid]

    if recursive:
        # Handles subgraphs.
        for i in range(len(new_nodes)):
            node = new_nodes[i]
            if node is None or not (node.attribute):
                continue
            new_nodes[i] = _apply_remove_node_fct_node(
                onnx_remove_node_unused,
                node,
                recursive=True,
                debug_info=debug_info + [node.name],
            )

    # Finally create the new graph.
    nodes = list(filter(lambda n: n is not None, new_nodes))
    if is_function:
        logger.debug("onnx_remove_node_unused:end function with %d nodes.", len(nodes))
        return make_function(
            onnx_model.domain,
            onnx_model.name,
            onnx_model.input,
            onnx_model.output,
            nodes,
            opset_imports=onnx_model.opset_import,
            attributes=onnx_model.attribute,
            doc_string=onnx_model.doc_string,
        )
    graph = make_graph(
        nodes, onnx_model.name, onnx_model.input, onnx_model.output, new_inits
    )
    graph.value_info.extend(onnx_model.value_info)
    logger.debug("onnx_remove_node_unused:end graph with %d nodes.", len(nodes))
    return graph


def _guess_proto_dtype(dtype) -> int:
    return np_dtype_to_tensor_dtype(dtype)


def get_tensor_shape(
    obj: Union[ValueInfoProto, TypeProto, TensorProto]
) -> Optional[List[Union[int, str, None]]]:
    """
    Returns the shape if that makes sense for this object.
    """
    if isinstance(obj, ValueInfoProto):
        return get_tensor_shape(obj.type)
    elif not isinstance(obj, TypeProto):
        raise TypeError(f"Unexpected type {type(obj)!r}.")
    if not obj.tensor_type.HasField("shape"):
        return None
    shape = []
    for d in obj.tensor_type.shape.dim:
        v = d.dim_value if d.dim_value > 0 else d.dim_param
        shape.append(v)
    if len(shape) == 0:
        return shape
    return list(None if s in (0, "") else s for s in shape)


def get_hidden_inputs(nodes: Iterable[NodeProto]) -> Set[str]:
    """
    Returns the list of hidden inputs used by subgraphs.

    :param nodes: list of nodes
    :return: list of names
    """
    inputs = set()
    outputs = set()
    for node in nodes:
        inputs |= set(node.input)
        outputs |= set(node.output)
        for att in node.attribute:
            if (
                att.type != AttributeProto.GRAPH
                or not hasattr(att, "g")
                or att.g is None
            ):
                continue
            hidden = get_hidden_inputs(att.g.node)
            inits = set([i.name for i in att.g.initializer])
            inputs |= hidden - (inits & hidden)
    return inputs - (outputs & inputs)


def enumerate_model_node_outputs(
    model: ModelProto, add_node: bool = False, order: bool = False
) -> Iterable:
    """
    Enumerates all the nodes of a model.

    :param model: :epkg:`ONNX` graph
    :param add_node: if False, the function enumerates
        all output names from every node, otherwise, it
        enumerates tuple (output name, node)
    :param order: goes through outputs following the graph order
    :return: enumerator
    """
    if not hasattr(model, "graph"):
        raise TypeError(f"Parameter model is not an ONNX model but {type(model)}")
    if order:
        edges = []
        dorder = {}
        node_names = {}
        for inp in model.graph.input:
            dorder[0, inp.name] = 0
        for node in model.graph.node:
            dorder[1, node.name] = 0
            for i in node.input:
                edges.append(("in", i, node.name))
            for o in node.output:
                edges.append(("out", o, node.name))
                node_names[o] = node
                dorder[0, o] = 0

        modif = 1
        n_iter = 0
        while modif > 0 and n_iter <= len(model.graph.node):
            modif = 0
            n_iter += 1
            for kind, data_name, node_name in edges:
                if kind == "in":
                    if (0, data_name) not in dorder:
                        continue
                    if dorder[0, data_name] + 1 > dorder[1, node_name]:
                        modif += 1
                        dorder[1, node_name] = dorder[0, data_name] + 1
                else:
                    if dorder[1, node_name] + 1 > dorder[0, data_name]:
                        modif += 1
                        dorder[0, data_name] = dorder[1, node_name] + 1

        orders = [(v, k) for k, v in dorder.items()]
        orders.sort()

        for _, k in orders:
            if k[0] == 1:
                continue
            out = k[1]
            if out not in node_names:
                continue
            yield (out, node_names[out]) if add_node else out
    else:
        for node in model.graph.node:
            for out in node.output:
                yield (out, node) if add_node else out


def select_model_inputs_outputs(
    model: ModelProto,
    outputs: Optional[List[str]] = None,
    inputs: Optional[List[str]] = None,
    infer_shapes: bool = True,
    overwrite: Optional[Dict[str, Any]] = None,
    remove_unused: bool = True,
    verbose: int = 0,
):
    """
    Takes a model and changes its outputs.

    :param model: :epkg:`ONNX` model
    :param inputs: new inputs, same ones if None
    :param outputs: new outputs, same ones if None
    :param infer_shapes: infer inputs and outputs shapes
    :param overwrite: overwrite type and shapes for
        inputs or outputs, *overwrite* is a
        dictionary `{'name': (numpy dtype, shape)}`
    :param remove_unused: remove unused nodes from the graph
    :param verbose: display information while converting
    :return: modified model

    The function removes unneeded nodes.

    The following example shows how to change the inputs of model
    to bypass the first nodes. Shape inferences fails to determine
    the new inputs type. They need to be overwritten.
    `verbose=1` shows the number of deleted nodes.

    ::

        import onnx
        from onnx_extended.tools.onnx_manipulations import select_model_inputs_outputs

        onx = onnx.load(path)
        onx2 = select_model_inputs_outputs(
            onx, inputs=["a", "b"],
            infer_shapes=True, verbose=1,
            overwrite={'a': (numpy.int32, None), 'b': (numpy.int64, None)})
        onnx.save(onx2, path2)
    """
    if inputs is not None and not isinstance(inputs, list):
        inputs = [inputs]
    if outputs is not None and not isinstance(outputs, list):
        outputs = [outputs]
    if inputs is None:
        inputs = [i.name for i in model.graph.input]
    if outputs is None:
        outputs = [o.name for o in model.graph.output]

    mark_var = {}
    for out in enumerate_model_node_outputs(model):
        mark_var[out] = 0
    for inp in inputs:
        mark_var[inp] = 0
    for out in outputs:
        if out not in mark_var:
            raise ValueError(f"Output '{out}' not found in model.")
        mark_var[out] = 1

    nodes = list(model.graph.node[::-1])
    mark_op = {}
    for node in list(nodes):
        mark_op[id(node)] = 0

    # We mark all the nodes we need to keep.
    nb = 1
    while nb > 0:
        nb = 0
        for node in nodes:
            if mark_op[id(node)] == 1:
                continue
            mod = False
            for out in node.output:
                if mark_var[out] == 1:
                    mark_op[id(node)] = 1
                    mod = True
                    break
            if not mod:
                continue

            hidden = get_hidden_inputs([node])
            node_inputs = list(node.input) + list(hidden)

            nb += 1
            for inp in node_inputs:
                if inp in inputs:
                    continue
                if mark_var.get(inp, 0) == 1:
                    continue
                mark_var[inp] = 1
                nb += 1

    # All nodes verifies mark_op[node.name] == 1
    keep_nodes = [node for node in nodes[::-1] if mark_op[id(node)] == 1]

    if verbose > 1:
        for node in nodes:
            s = "+" if mark_op[id(node)] == 1 else "-"
            logger.info(
                "[select_model_inputs_outputs] %s %s (%s) -> %s [%s]"
                % (
                    s,
                    node.op_type,
                    ", ".join(node.input),
                    ", ".join(node.output),
                    node.name,
                )
            )

    known_shapes = {}
    if infer_shapes:
        shapes = shape_inference.infer_shapes(model)
        for shape in shapes.graph.value_info:
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.input:
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.output:
            known_shapes[shape.name] = shape.type
    else:
        for shape in model.graph.input:
            known_shapes[shape.name] = shape.type
        for shape in model.graph.output:
            known_shapes[shape.name] = shape.type

    var_in = []
    for name in inputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = _guess_proto_dtype(dtype)
            value_info = make_tensor_value_info(name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = ValueInfoProto()
                value_info.name = name
            else:
                shape = get_tensor_shape(known_shapes[name])
                value_info = make_tensor_value_info(name, proto_dtype, shape)
        else:
            value_info = ValueInfoProto()
            value_info.name = name
        var_in.append(value_info)

    var_out = []
    for name in outputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = _guess_proto_dtype(dtype)
            value_info = make_tensor_value_info(name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = ValueInfoProto()
                value_info.name = name
            else:
                shape = get_tensor_shape(known_shapes[name])
                value_info = make_tensor_value_info(name, proto_dtype, shape)
        else:
            value_info = ValueInfoProto()
            value_info.name = name
        var_out.append(value_info)

    if verbose > 0:
        logger.info(
            "[select_model_inputs_outputs] nodes %r --> %r"
            % (len(model.graph.node), len(keep_nodes))
        )
        logger.info(
            "[select_model_inputs_outputs] inputs: %r" % [_.name for _ in var_in]
        )
        logger.info(
            "[select_model_inputs_outputs] inputs: %r" % [_.name for _ in var_out]
        )

    graph = make_graph(
        keep_nodes,
        model.graph.name,
        var_in,
        var_out,
        model.graph.initializer,
        sparse_initializer=model.graph.sparse_initializer,
    )
    onnx_model = make_model(graph, functions=model.functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()
        op_set.domain = oimp.domain
        op_set.version = oimp.version

    # remove unused nodes
    if remove_unused:
        onnx_model = onnx_remove_node_unused(onnx_model, recursive=False)

    return onnx_model



import json
import warnings
from typing import List, Optional, Union
from pandas import DataFrame


_mapping_types = {
    "float": "F",
    "double": "D",
    "float16": "H",
    "uint8": "U8",
    "uint16": "U16",
    "uint32": "U32",
    "uint64": "U64",
    "int8": "I8",
    "int16": "I16",
    "int32": "I32",
    "int64": "I64",
}


def _process_shape(shape_df):
    if isinstance(shape_df, float) or len(shape_df) == 0:
        return ""
    values = []
    for val in shape_df:
        if len(val) != 1:
            raise ValueError(f"Unable to process shape {val!r} from {values!r}.")
        k, v = list(val.items())[0]
        if len(v) > 0:
            vs = "x".join(map(str, v))
            values.append(f"{_mapping_types.get(k,k)}[{vs}]")
        else:
            values.append(f"{_mapping_types.get(k,k)}")
    return "+".join(values)


def post_process_df_profile(
    df: DataFrame,
    first_it_out: bool = False,
    agg: bool = False,
    agg_op_name: bool = True,
    with_shape: bool = False,
) -> DataFrame:
    """
    Post-processed a dataframe obtained after profiling onnxruntime.
    It adds a column for a more explicit event name and adds
    a column for the iteration number

    :param agg: aggregate the result
    :param first_it_out: leave the first iteration
        out of the aggregation
    :param agg_op_name: aggregate on operator name or operator index
    :param with_shape: keep the shape to aggregate
    :return: DataFrame
    """
    events = {"kernel_time", "fence_after", "fence_before"}

    def sep_event(s):
        for e in events:
            if s.endswith(e):
                return e
        return s

    df = df.copy()
    df["event_name"] = df["name"].apply(sep_event)
    df["iteration"] = -1
    current = -1
    for i in range(df.shape[0]):
        if df.loc[i, "name"] == "SequentialExecutor::Execute":
            current += 1
        df.loc[i, "iteration"] = current

    if not agg:
        if with_shape:
            df["args_input_type_shape"] = df["args_input_type_shape"].apply(
                _process_shape
            )
            df["args_output_type_shape"] = df["args_output_type_shape"].apply(
                _process_shape
            )
        else:
            df = df.drop(["args_input_type_shape", "args_output_type_shape"], axis=1)
        if first_it_out:
            df["it==0"] = (df["iteration"] <= 0).astype(int)
        return df

    agg_cols = ["cat", "args_node_index", "args_op_name", "args_provider", "event_name"]
    if with_shape:
        agg_cols.append("args_input_type_shape")
        df["args_input_type_shape"] = df["args_input_type_shape"].apply(_process_shape)
        df["args_output_type_shape"] = df["args_output_type_shape"].apply(
            _process_shape
        )
    else:
        df = df.drop(["args_input_type_shape", "args_output_type_shape"], axis=1)

    if first_it_out:
        df["it==0"] = (df["iteration"] <= 0).astype(int)
        agg_cols.insert(0, "it==0")
    if agg_op_name:
        del agg_cols[agg_cols.index("args_node_index")]
    for c in agg_cols:
        df[c] = df[c].fillna("")
    df["dur"] = df["dur"].fillna(0)
    agg = df[agg_cols + ["dur"]].groupby(agg_cols).sum()
    return agg


def js_profile_to_dataframe(
    filename: str,
    as_df: bool = True,
    first_it_out: bool = False,
    agg: bool = False,
    agg_op_name: bool = False,
    with_shape: bool = False,
) -> Union[List, DataFrame]:
    """
    Profiles the execution of an onnx graph with onnxruntime.

    :param filename: filename holding the profiling stored in json format
    :param as_df: returns the
    :param first_it_out: if aggregated, leaves the first iteration out
    :param agg: aggregate by event
    :param agg_op_name: aggregate on operator name or operator index
    :param with_shape: keep the shape before aggregating
    :return: DataFrame or dictionary
    """
    with open(filename, "r") as f:
        content = f.read()
    js = json.loads(content)

    suffixes = ["_kernel_time", "_fence_before", "_fence_after"]
    rows = []
    for row in js:
        if "args" in row and isinstance(row["args"], dict):
            for k, v in row["args"].items():
                row[f"args_{k}"] = v
            del row["args"]
        name = row["name"]
        for suf in suffixes:
            if name.endswith(suf):
                changed = name[: -len(suf)]
                row["op_name"] = changed
                break
        rows.append(row)
    if as_df:
        return post_process_df_profile(
            DataFrame(rows),
            first_it_out=first_it_out,
            agg=agg,
            agg_op_name=agg_op_name,
            with_shape=with_shape,
        )
    return rows


def _preprocess_graph1(df):
    df = df.copy()
    df["args_provider"] = df["args_provider"].apply(
        lambda s: s.replace("ExecutionProvider", "") if isinstance(s, str) else s
    )
    agg_cols = ["dur", "args_op_name", "args_provider"]
    for c in ["it==0", "args_input_type_shape"]:
        if c in df.columns:
            agg_cols.append(c)
    gr_dur = df[agg_cols].groupby(agg_cols[1:]).sum().sort_values("dur")
    gr_n = df[agg_cols].groupby(agg_cols[1:]).count()
    gr_n = gr_n.loc[gr_dur.index, :]
    gr_n.columns = ["count"]
    gr = gr_dur.merge(gr_n, left_index=True, right_index=True, how="outer")
    gr["ratio"] = gr["dur"] / gr["dur"].sum()
    return gr_dur, gr_n, gr


def _preprocess_graph2(df):
    df = df.reset_index(drop=False).copy()
    df["args_node_index"] = df["args_node_index"].apply(
        lambda i: int(i) if i not in {None, ""} else -1
    )
    df["args_provider"] = df["args_provider"].apply(
        lambda s: s.replace("ExecutionProvider", "") if isinstance(s, str) else s
    )
    df = df[(df["cat"] == "Node") & (df["event_name"] == "kernel_time")]
    agg_cols = ["dur", "args_node_index", "args_op_name", "args_provider"]
    for c in ["it==0", "args_input_type_shape"]:
        if c in df.columns:
            agg_cols.append(c)
    df = df[agg_cols].groupby(agg_cols[1:]).sum()
    df = df.sort_index(ascending=False)
    df["ratio"] = df["dur"] / df["dur"].sum()
    return df


def plot_ort_profile(
    df: DataFrame,
    ax0: Optional["matplotlib.axes.Axes"] = None,
    ax1: Optional["matplotlib.axes.Axes"] = None,
    title: Optional[str] = None,
) -> "matplotlib.axes.Axes":
    """
    Plots time spend in computation based on dataframe
    produced by function :func:`js_profile_to_dataframe`.

    :param df: dataframe
    :param ax0: first axis to draw time
    :param ax1: second axis to draw occurences
    :param title: graph title
    :return: the graph
    """
    fontsize = 10
    if ax0 is None:
        import matplotlib as plt

        ax0 = plt.gca()  # pragma: no cover

    if "args_provider" in df.columns:
        # Aggregation by operator
        gr_dur, gr_n, _ = _preprocess_graph1(df)
        gr_dur.plot.barh(ax=ax0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ax0.set_xticklabels(ax0.get_xticklabels(), fontsize=fontsize)
            ax0.get_yaxis().set_label_text("")
            ax0.set_yticklabels(
                ax0.get_yticklabels(), rotation=45, ha="right", fontsize=fontsize
            )
        if title is not None:
            ax0.set_title(title)
        if ax1 is not None:
            gr_n.plot.barh(ax=ax1)
            ax1.set_title("n occurences")
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                ax1.set_xticklabels(ax1.get_xticklabels(), fontsize=fontsize)
                ax1.get_yaxis().set_label_text("")
                ax1.set_yticklabels(
                    ax1.get_yticklabels(), rotation=45, ha="right", fontsize=fontsize
                )
        return ax0

    df = _preprocess_graph2(df)
    df[["dur"]].plot.barh(ax=ax0)
    if title is not None:
        ax0.set_title(title)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ax0.set_xticklabels(ax0.get_xticklabels(), fontsize=fontsize)
        ax0.get_yaxis().set_label_text("")
        ax0.set_yticklabels(ax0.get_yticklabels(), fontsize=fontsize)
    return ax0



from typing import Dict, Optional
from onnx import AttributeProto, FunctionProto, TensorProto
from onnx.helper import (
    make_function,
    make_node,
    make_opsetid,
    make_tensor,
)


def make_dynamic_quantize_linear_function_proto(
    domain: str, opset: int, to: Optional[int] = None
) -> FunctionProto:
    """
    Creates the FunctionProto for function `DynamicQuantizeLinear`
    doing a quantization to float 8.

    :param domain: local domain name
    :param opset: opset to use to define the function
    :param to: if None, the function has an attribute,
        otherwise, it is replaced by the given value
    :return: FunctionProto

    The function takes 1 input and returns 3 outputs like
    operator `DynamicQuantizeLinear
    <https://onnx.ai/onnx/operators/onnx__DynamicQuantizeLinear.html>`_.
    It has one attribute *to* which specified the quantized type.
    """
    normalization_values = list(
        {
            TensorProto.FLOAT8E4M3FN: 100.057724,
            TensorProto.FLOAT8E4M3FNUZ: 54.26635,
            TensorProto.FLOAT8E5M2: 9535.286,
            TensorProto.FLOAT8E5M2FNUZ: 9403.499,
        }.items()
    )

    if to is None:
        cast = make_node("Cast", ["zerof"], ["Zeropoint"])
        att = AttributeProto()
        att.name = "to"
        att.ref_attr_name = "to"
        att.type = AttributeProto.INT
        cast.attribute.append(att)

        cst = make_node("Constant", [], ["vto"])
        att = AttributeProto()
        att.name = "value_int"
        att.ref_attr_name = "to"
        att.type = AttributeProto.INT
        cst.attribute.append(att)
    else:
        cast = make_node("Cast", ["zerof"], ["Zeropoint"], to=to)
        cst = make_node("Constant", [], ["vto"], value_int=to)

    nodes = [
        make_node(
            "Constant",
            [],
            ["zerof"],
            value=make_tensor("zerof", TensorProto.FLOAT, [], [0]),
        ),
        make_node(
            "Constant",
            [],
            ["newshape"],
            value=make_tensor("newshape", TensorProto.INT64, [1], [-1]),
        ),
        make_node("CastLike", ["zerof", "x"], ["zero"]),
        cast,
        make_node("IsNaN", ["x"], ["nanxp"]),
        make_node("Not", ["nanxp"], ["nanx"]),
        make_node("CastLike", ["nanx", "x"], ["nanxc"]),
        make_node("Where", ["nanx", "x", "zero"], ["xf"]),
        make_node("Mul", ["xf", "xf"], ["xsquare"]),
        make_node("ReduceSum", ["xsquare"], ["Num"], keepdims=0),
        make_node("ReduceSum", ["nanxc"], ["Den"], keepdims=0),
        make_node("Div", ["Num", "Den"], ["Dev"]),
        make_node("Sqrt", ["Dev"], ["Scale"]),
        cst,
        make_node("Reshape", ["vto", "newshape"], ["vtotensor"]),
        make_node(
            "LabelEncoder",
            ["vtotensor"],
            ["stdftensor"],
            keys_int64s=[v[0] for v in normalization_values],
            values_floats=[v[1] for v in normalization_values],
            domain="ai.onnx.ml",
        ),
        make_node("ReduceSum", ["stdftensor"], ["stdf"], keepdims=0),
        make_node("CastLike", ["stdf", "Scale"], ["std"]),
        make_node("Div", ["Scale", "std"], ["ScaleScaled"]),
        make_node("QuantizeLinear", ["x", "ScaleScaled", "Zeropoint"], ["y"]),
    ]
    return make_function(
        domain,
        "DynamicQuantizeLinear",
        ["x"],
        ["y", "ScaleScaled", "Zeropoint"],
        nodes,
        opset_imports=[make_opsetid("", opset), make_opsetid("ai.onnx.ml", 2)],
        attributes=["to"],
    )


def make_simple_dynamic_quantize_linear_function_proto(
    domain: str, opset: int, to: int = TensorProto.FLOAT8E4M3FN
) -> FunctionProto:
    """
    Creates the FunctionProto for function `SimpleDynamicQuantizeLinear`
    doing a quantization to float 8. A suffix is added to the function name
    to tell which type is used for the quantization. It does not
    support nan values.

    :param domain: local domain name
    :param opset: opset to use to define the function
    :param to: type to quantize into, it is hardcoded
    :return: FunctionProto

    The function takes 1 input and returns 3 outputs like
    operator `DynamicQuantizeLinear
    <https://onnx.ai/onnx/operators/onnx__DynamicQuantizeLinear.html>`_.
    It has one attribute *to* which specified the quantized type.
    """
    normalization_values: Dict[int, float] = {
        TensorProto.FLOAT8E4M3FN: 100.057724,
        TensorProto.FLOAT8E4M3FNUZ: 54.26635,
        TensorProto.FLOAT8E5M2: 9535.286,
        TensorProto.FLOAT8E5M2FNUZ: 9403.499,
    }
    suffix: Dict[int, str] = {
        TensorProto.FLOAT8E4M3FN: "E4M3FN",
        TensorProto.FLOAT8E4M3FNUZ: "E4M3FNUZ",
        TensorProto.FLOAT8E5M2: "E5M2",
        TensorProto.FLOAT8E5M2FNUZ: "E5M2FNUZ",
    }

    nodes = [
        make_node(
            "Constant",
            [],
            ["zerof"],
            value=make_tensor("zerof", TensorProto.FLOAT, [], [0]),
        ),
        make_node("Cast", ["zerof"], ["Zeropoint"], to=to),
        make_node(
            "Constant",
            [],
            ["stdf"],
            value=make_tensor(
                "stdf", TensorProto.FLOAT, [], [normalization_values[to]]
            ),
        ),
        make_node(
            "Constant",
            [],
            ["newshape"],
            value=make_tensor("newshape", TensorProto.INT64, [1], [-1]),
        ),
        make_node("Mul", ["x", "x"], ["xsquare"]),
        make_node("ReduceMean", ["xsquare"], ["Dev"], keepdims=0),
        make_node("Sqrt", ["Dev"], ["Scale"]),
        make_node("CastLike", ["stdf", "Scale"], ["std"]),
        make_node("Div", ["Scale", "std"], ["ScaleScaled"]),
        make_node("QuantizeLinear", ["x", "ScaleScaled", "Zeropoint"], ["y"]),
    ]
    return make_function(
        domain,
        f"DynamicQuantizeLinear{suffix[to]}",
        ["x"],
        ["y", "ScaleScaled", "Zeropoint"],
        nodes,
        opset_imports=[make_opsetid("", opset)],
        attributes=["to"],
    )



from typing import Any, Dict, Iterator, List, Union, Tuple
import numpy as np
from onnx import AttributeProto, ModelProto, NodeProto, load
from onnx.reference.op_run import to_array_extended
from onnxruntime import InferenceSession
from .onnx_manipulations import select_model_inputs_outputs


def render_node(node: NodeProto) -> str:
    """
    Renders a node into text to display it.

    :param node: Node
    :return: trext
    """
    att_display = [
        "activations",
        "align_corners",
        "allowzero",
        "alpha",
        "auto_pad",
        "axis",
        "axes",
        "batch_axis",
        "batch_dims",
        "beta",
        "bias",
        "blocksize",
        "case_change_action",
        "ceil_mode",
        "center_point_box",
        "clip",
        "coordinate_transformation_mode",
        "count_include_pad",
        "cubic_coeff_a",
        "decay_factor",
        "detect_negative",
        "detect_positive",
        "dilation",
        "dilations",
        "direction",
        "dtype",
        "end",
        "epsilon",
        "equation",
        "exclusive",
        "exclude_outside",
        "extrapolation_value",
        "fmod",
        "gamma",
        "group",
        "hidden_size",
        "high",
        "ignore_index",
        "input_forget",
        "is_case_sensitive",
        "k",
        "keepdims",
        "kernel_shape",
        "lambd",
        "largest",
        "layout",
        "linear_before_reset",
        "locale",
        "low",
        "max_gram_length",
        "max_skip_count",
        "mean",
        "min_gram_length",
        "mode",
        "momentum",
        "nearest_mode",
        "ngram_counts",
        "ngram_indexes",
        "noop_with_empty_axes",
        "norm_coefficient",
        "norm_coefficient_post",
        "num_scan_inputs",
        "output_height",
        "output_padding",
        "output_shape",
        "output_width",
        "p",
        "padding_mode",
        "pads",
        "perm",
        "pooled_shape",
        "reduction",
        "reverse",
        "sample_size",
        "sampling_ratio",
        "scale",
        "scan_input_axes",
        "scan_input_directions",
        "scan_output_axes",
        "scan_output_directions",
        "seed",
        "select_last_index",
        "size",
        "sorted",
        "spatial_scale",
        "start",
        "storage_order",
        "strides",
        "time_axis",
        "to",
        "training_mode",
        "transA",
        "transB",
        "type",
        "upper",
        "xs",
        "y",
        "zs",
    ]

    sub_graphs_names: Dict[str, str] = {}

    def _get_subgraph_name(idg):
        if idg in sub_graphs_names:
            return sub_graphs_names[idg]
        g = "G%d" % (len(sub_graphs_names) + 1)
        sub_graphs_names[idg] = g
        return g

    def str_node(indent, node):
        atts = []
        if hasattr(node, "attribute"):
            for att in node.attribute:
                done = True
                if hasattr(att, "ref_attr_name") and att.ref_attr_name:
                    atts.append(f"{att.name}=${att.ref_attr_name}")
                    continue
                if att.name in att_display:
                    if att.type == AttributeProto.INT:
                        atts.append("%s=%d" % (att.name, att.i))
                    elif att.type == AttributeProto.FLOAT:
                        atts.append(f"{att.name}={att.f:1.2f}")
                    elif att.type == AttributeProto.INTS:
                        atts.append(
                            "%s=%s" % (att.name, str(list(att.ints)).replace(" ", ""))
                        )
                    else:
                        done = False
                elif (
                    att.type == AttributeProto.GRAPH
                    and hasattr(att, "g")
                    and att.g is not None
                ):
                    atts.append(f"{att.name}={_get_subgraph_name(id(att.g))}")
                else:
                    done = False
                if done:
                    continue
                if att.type in (
                    AttributeProto.TENSOR,
                    AttributeProto.TENSORS,
                    AttributeProto.SPARSE_TENSOR,
                    AttributeProto.SPARSE_TENSORS,
                ):
                    try:
                        val = str(to_array_extended(att.t).tolist())
                    except TypeError as e:
                        raise TypeError(  # pragma: no cover
                            "Unable to display tensor type %r.\n%s"
                            % (att.type, str(att))
                        ) from e
                    if "\n" in val:
                        val = val.split("\n", maxsplit=1) + "..."
                    if len(val) > 10:
                        val = val[:10] + "..."
                elif att.type == AttributeProto.STRING:
                    val = str(att.s)
                    if len(val) > 50:
                        val = val[:40] + "..." + val[-10:]
                elif att.type == AttributeProto.STRINGS:
                    n_val = list(att.strings)
                    if len(n_val) < 5:
                        val = ",".join(map(str, n_val))
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:2])),
                            ",".join(map(str, n_val[-2:])),
                        )
                elif att.type == AttributeProto.INT:
                    val = str(att.i)
                elif att.type == AttributeProto.FLOAT:
                    val = str(att.f)
                elif att.type == AttributeProto.INTS:
                    n_val = list(att.ints)
                    if len(n_val) < 6:
                        val = f"[{','.join(map(str, n_val))}]"
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:3])),
                            ",".join(map(str, n_val[-3:])),
                        )
                elif att.type == AttributeProto.FLOATS:
                    n_val = list(att.floats)
                    if len(n_val) < 5:
                        val = f"[{','.join(map(str, n_val))}]"
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:2])),
                            ",".join(map(str, n_val[-2:])),
                        )
                else:
                    val = ".%d" % att.type
                atts.append(f"{att.name}={val}")
        inputs = list(node.input)
        if len(atts) > 0:
            inputs.extend(atts)
        domain = "" if node.domain in ("", "ai.onnx.ml") else f"[{node.domain}]"
        return "%s%s%s(%s) -> %s" % (
            "  " * indent,
            node.op_type,
            domain,
            ", ".join(inputs),
            ", ".join(node.output),
        )

    return str_node(0, node)


def enumerate_ort_run(
    onx: Union[str, ModelProto], feeds: Dict[str, Any], verbose: int = 0
) -> Iterator[Tuple[List[str], List[Any]]]:
    """
    Yields all the intermediate results produced by
    :epkg:`onnxruntime`.

    :param onx: model
    :param feeds: input tensors
    :param verbose: prints out a summary of the results
    :return: intermediate results and names
    """
    if isinstance(onx, str):
        with open(onx, "rb") as f:
            proto = load(f)
    else:
        proto = onx

    inputs = [i.name for i in proto.graph.input]
    if verbose == 1:
        import tqdm

        loop = tqdm.tqdm(proto.graph.node)
    else:
        loop = proto.graph.node
    if verbose > 1:
        for init in proto.graph.initializer:
            value = to_array_extended(init)
            if verbose <= 2:
                print(" + %s: %s%s" % (init.name, value.dtype, value.shape))
            else:
                print(" + %s: %s" % (init.name, value))
        for i in onx.graph.input:
            if i.name not in feeds:
                continue
            value = feeds[i.name]
            if verbose <= 2:
                print(" + %s: %s%s" % (i.name, value.dtype, value.shape))
            else:
                print(" + %s: %s" % (i.name, value))
    for node in loop:
        names = list(node.output)
        if verbose > 1:
            print(render_node(node))
        subproto = select_model_inputs_outputs(proto, outputs=names, inputs=inputs)

        sess = InferenceSession(
            subproto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        outputs = sess.run(None, feeds)
        if verbose > 1:
            for name, value in zip(node.output, outputs):
                if isinstance(value, np.ndarray) and verbose <= 2:
                    print(" + %s: %s%s" % (name, value.dtype, value.shape))
                else:
                    print(" + %s: %s%s" % (name, value.dtype, value.shape))
                    print(value)
        yield names, outputs



from enum import IntEnum
from logging import getLogger
from typing import Dict, List, Optional, Tuple, Union
import numpy as np
from onnx import FunctionProto, NodeProto, TensorProto
from onnx.helper import (
    make_node,
    make_tensor,
    make_tensor_value_info,
    tensor_dtype_to_np_dtype,
)
from onnx.numpy_helper import float8e4m3_to_float32, float8e5m2_to_float32
from onnx.reference.custom_element_types import float8e4m3fn
from onnx.reference.op_run import to_array_extended
from onnx.onnx_cpp2py_export.defs import SchemaError

try:
    from onnx.reference.ops.op_cast import Cast_19 as Cast
    from onnx.reference.ops.op_quantize_linear import (
        QuantizeLinear_19 as QuantizeLinear,
    )
except ImportError:
    from onnx.reference.ops.op_cast import Cast
    from onnx.reference.ops.op_quantize_linear import QuantizeLinear
from ...helper import (
    make_dynamic_quantize_linear_function_proto,
    make_matmul_reshape_transpose_back_function_proto,
    make_matmul_reshape_transpose_function_proto,
    make_simple_dynamic_quantize_linear_function_proto,
)
from ...reference import from_array_extended
from ...validation.cython.fp8 import cast_float32_to_e4m3fn
from .errors import QuantizationError
from .onnx_graph_struct import _get_shape, Graph, Node, NodeKind


logger = getLogger("onnx-extended/transformer")


class QuantizeOptions(IntEnum):
    """
    Quantization options.

    * `NONE`: no option
    * `OPTIMIZE`: assumes there is no nan values,
        choose less generic functions such as the implementation
        of DynamicQuantizeLinear
    """

    NONE = 0
    OPTIMIZE = 1


class TransformResults:
    """
    Output of a function transforming a graph.

    :param removed_nodes: node to remove from the graph
    :param added_nodes: node to add to the graph
    :param new_opsets: opsets to update
    :param local_functions: necessary functions to add to the graph
    """

    def __init__(
        self,
        removed_nodes: List[Node],
        added_nodes: List[NodeProto],
        new_opsets: Optional[Dict[str, int]] = None,
        local_functions: Optional[List[FunctionProto]] = None,
    ):
        self.removed_nodes = removed_nodes
        self.added_nodes = added_nodes
        self.new_opsets = new_opsets or {}
        self.local_functions = local_functions or []


def estimation_quantization_scale(
    coef: np.array,
    to: int = TensorProto.FLOAT8E4M3FN,
    threshold: float = 0.99999,
) -> Tuple[float, float]:
    """
    Estimates the scale parameter for the quantization to float 8 assuming
    the distribution of the coefficients is gaussian.
    """

    if to in (
        TensorProto.FLOAT8E4M3FN,
        TensorProto.FLOAT8E4M3FNUZ,
        TensorProto.FLOAT8E5M2,
        TensorProto.FLOAT8E5M2FNUZ,
    ):
        if to == TensorProto.FLOAT8E4M3FN:
            fct = float8e4m3_to_float32
        elif to == TensorProto.FLOAT8E4M3FNUZ:

            def fct(x):
                return float8e4m3_to_float32(x, uz=True)

        elif to == TensorProto.FLOAT8E5M2:
            fct = float8e5m2_to_float32
        elif to == TensorProto.FLOAT8E5M2FNUZ:

            def fct(x):
                return float8e5m2_to_float32(x, uz=True, fn=True)

        else:
            raise ValueError(f"Unexpected to={to!r}.")

        float8 = [fct(i) for i in range(0, 256)]
        quant_float = [f for f in float8 if not np.isnan(f) and not np.isinf(f)]
        std_coef = np.mean(coef.ravel() ** 2) ** 0.5
        std_quant = np.std(np.array(quant_float, dtype=np.float32))
        zero = 0.0
        scale = std_quant.astype(coef.dtype) / std_coef.astype(coef.dtype)
    elif to == TensorProto.UINT8:
        qu = np.quantile(coef.ravel(), [1 - threshold, threshold])
        scale = 255 / (qu[1] - qu[0])
        zero = qu[0] * scale
    elif to == TensorProto.INT8:
        qu = np.quantile(coef.ravel(), [1 - threshold, threshold])
        scale = 254 / (qu[1] - qu[0])
        zero = (qu[0] + qu[1]) / 2 * scale
    else:
        raise ValueError(f"Unexpected quantization type for to={to}.")

    return np.array(1.0 / scale, dtype=coef.dtype), -zero


def quantize_weights(
    node: Node,
    elem_type: int,
    index: int,
    transpose: bool = False,
) -> Tuple[Node, Node, Node, Union[bool, Tuple[int, ...]], Tuple[int, ...]]:
    """
    Quantizes a tensor into a tensor of element type *elem_type*.

    :param node: Node to quantize
    :param elem_type: element type
    :param transpose: transpose the weight before doing it
    :param verbose: logs what is quantized
    :return: three new nodes, quantized weights, scale, zero point
        and the original shape of the constant, shape

    See function :func:`quantize_float8_matmul`
    """
    tensor = node.get_tensor()
    values = to_array_extended(tensor)
    shape = values.shape
    if len(values.shape) == 1:
        raise NotImplementedError(
            f"Input {index} is a constant, it must have at "
            f"least 2 dimensions not {values.shape}."
        )
    if len(values.shape) > 2:
        # Needs to be reshaped.
        reshaped = values.shape
        if index == 0:
            # first input
            values = values.reshape((-1, values.shape[-1]))
        else:
            # second input
            values = (
                values.reshape((-1,) + values.shape[-2:])
                .transpose((1, 0, 2))
                .reshape((values.shape[-2], -1))
            )
    else:
        reshaped = False
    if transpose:
        values = values.T
    scale, zp = estimation_quantization_scale(values, to=elem_type)
    zpt = make_tensor("zp", elem_type, [], [zp])
    zpa = to_array_extended(zpt)

    if elem_type == TensorProto.FLOAT8E4M3FN:
        new_values = cast_float32_to_e4m3fn(values / scale).astype(float8e4m3fn)
    else:
        new_values = QuantizeLinear.eval(values, scale, zpa)

    new_tensor = from_array_extended(
        new_values, name=node.parent.generate_name(node.outname)
    )
    node_weight = node.create_with_new_values(new_tensor)

    new_scale = from_array_extended(
        scale, name=node.parent.generate_name(f"{node.outname}_scale")
    )
    node_scale = node.create_with_new_values(new_scale)

    new_zp = from_array_extended(
        zpa, name=node.parent.generate_name(f"{node.outname}_zp")
    )
    node_zp = node.create_with_new_values(new_zp)

    return node_weight, node_scale, node_zp, reshaped, shape


class _MainQuantizeState:
    def __init__(self, version, domain_ops, local_functions, opset, index_transpose):
        self.domain_dq = "local.quant.domain"
        if domain_ops is None:
            domain_ops = {}
        self.domain_ops = domain_ops
        if version == "onnxruntime":
            self.op_gemm = "GemmFloat8"
            self.domain_gemm = domain_ops.get("GemmFloat8", "com.microsoft")
        elif version == "onnx-extended":
            self.op_gemm = "CustomGemmFloat8E4M3FN"
            self.domain_gemm = domain_ops.get(
                "CustomGemmFloat8E4M3FN", "onnx_extented.ortops.tutorial.cuda"
            )
        else:
            raise ValueError(f"Unexpected value {version!r} for version.")
        if local_functions is None:
            raise ValueError("local_functions cannot be None")
        self.local_functions = local_functions
        self.opset = opset
        self.index_transpose = index_transpose


class _QuantizeState:
    def __init__(self, logger, quantize_options, main_state):
        self.removed = []
        self.added = []
        self.input_names = []
        self.was_reshaped = [False, False]
        self.shapes = []
        self.logger = logger
        self.quantize_options = quantize_options
        self.main_state = main_state

    def quantize_connstant_weights(self, node, name, elem_type, index, do_transpose):
        cst = node.parent.get_node_producer(name)
        weight, scale, zero_point, reshaped, shape = quantize_weights(
            cst, elem_type, index, transpose=do_transpose
        )
        self.added.extend([weight.proto, scale.proto, zero_point.proto])
        self.input_names.append([weight.outname, scale.outname, zero_point.outname])
        self.removed.append(cst)
        self.was_reshaped[index] = reshaped
        self.shapes.append(shape)
        self.logger.debug("[quantize_weights] static quantize shape=%r", shape)

    def quantize_dynamic(self, node, name, elem_type, index, do_transpose):
        # Not a constant
        shape = node.parent.get_shape(name)
        self.shapes.append(shape)
        self.logger.debug("[quantize_float8_matmul] quantize dynamic shape %r", shape)
        if len(shape) == 2:
            if do_transpose:
                # no need
                temp_name = node.parent.generate_name(f"{name}_tr")
                self.added.append(
                    make_node(
                        "Transpose",
                        [name],
                        [temp_name],
                        perm=[1, 0],
                        name=node.parent.generate_node_name(f"tra8_{name}"),
                    )
                )
            else:
                temp_name = name
        else:
            if shape is None:
                raise QuantizationError(
                    f"Shape is unknown for result {name!r} in node {node}. "
                    f"This input cannot be transposed with certainty."
                )

            temp_name = node.parent.generate_name(f"{name}_tr")
            fname = f"MatMulReshapeTranspose{'T' if do_transpose else 'N'}{index}"
            self.added.append(
                make_node(
                    fname,
                    [name],
                    [temp_name],
                    perm=[1, 0],
                    domain=self.main_state.domain_dq,
                    name=node.parent.generate_node_name("MMRT"),
                )
            )
            self.was_reshaped[index] = name
            if (
                self.main_state.domain_dq,
                fname,
            ) not in self.main_state.local_functions:
                # use local functions
                self.main_state.local_functions[
                    self.main_state.domain_dq, fname
                ] = make_matmul_reshape_transpose_function_proto(
                    domain=self.main_state.domain_dq,
                    opset=self.main_state.opset,
                    index=index,
                    transpose=do_transpose,
                )

        new_name = node.parent.generate_name(f"{name}_f8")
        scale = node.parent.generate_name(f"{name}_scale")
        zero_point = node.parent.generate_name(f"{name}_zp")

        if self.quantize_options == QuantizeOptions.NONE:
            fname, fmake = (
                "DynamicQuantizeLinear",
                make_dynamic_quantize_linear_function_proto,
            )
        elif self.quantize_options == QuantizeOptions.OPTIMIZE:
            suffix = {
                TensorProto.FLOAT8E4M3FN: "E4M3FN",
                TensorProto.FLOAT8E4M3FNUZ: "E4M3FNUZ",
                TensorProto.FLOAT8E5M2: "E5M2",
                TensorProto.FLOAT8E5M2FNUZ: "E5M2FNUZ",
            }
            fname, fmake = (
                f"DynamicQuantizeLinear{suffix[elem_type]}",
                lambda domain=None, opset=None: (
                    make_simple_dynamic_quantize_linear_function_proto(
                        domain=domain, opset=opset, to=elem_type
                    )
                ),
            )
        else:
            raise RuntimeError(
                f"Unexpected value {self.quantize_options!r} for quantize_options."
            )
        proto = make_node(
            fname,
            [temp_name],
            [new_name, scale, zero_point],
            to=elem_type,
            domain=self.main_state.domain_dq,
            name=node.parent.generate_node_name("DQL"),
        )
        dql = Node(None, node.parent, proto, NodeKind.NODE)
        self.added.extend([dql.proto])
        self.input_names.append(dql.outputs)
        if (
            self.main_state.domain_dq == "local.quant.domain"
            and (self.main_state.domain_dq, fname)
            not in self.main_state.local_functions
        ):
            # use local functions
            self.main_state.local_functions[self.main_state.domain_dq, fname] = fmake(
                domain=self.main_state.domain_dq, opset=self.main_state.opset
            )

    def finalize(self, node, name, output_type):
        if self.was_reshaped[0] and self.was_reshaped[1]:
            raise QuantizationError(
                f"MatMul cannot be replaced by operator Gemm as both inputs "
                f"are not matrices. Their shapes are {self.shapes}. "
                f"Node name is {node.name!r}."
            )

        if output_type in {
            TensorProto.INT8,
            TensorProto.UINT8,
            TensorProto.FLOAT8E4M3FN,
            TensorProto.FLOAT8E4M3FNUZ,
            TensorProto.FLOAT8E5M2,
            TensorProto.FLOAT8E5M2FNUZ,
        }:
            # output is quantized, there is a need for a scale
            scale_out = node.parent.generate_name(f"{name}_scaleout")
            self.added.append(
                make_node(
                    "Mul",
                    [self.input_names[0][1], self.input_names[1][1]],
                    [scale_out],
                    name=node.parent.generate_node_name(f"mul8_{name}"),
                )
            )
        else:
            # output is not quantized, no need for an output scale
            scale_out = ""

        gemm_inputs = [
            self.input_names[0][0],  # A
            self.input_names[1][0],  # B
            "",  # C
            self.input_names[0][1],  # scaleA
            self.input_names[1][1],  # scaleB
            scale_out,  # scaleR
        ]
        if self.was_reshaped[0] or self.was_reshaped[1]:
            gemm_outputs = [node.parent.generate_name(f"{name}_gemm")]
            do_reshape = True
        else:
            gemm_outputs = node.outputs
            do_reshape = False

        self.added.append(
            make_node(
                self.main_state.op_gemm,
                gemm_inputs,
                gemm_outputs,
                rowMajor=0,
                dtype=output_type,
                transA=1 if (self.main_state.index_transpose & 1) else 0,
                transB=1 if (self.main_state.index_transpose & 2) else 0,
                domain=self.main_state.domain_gemm,
                computeType="CUBLAS_COMPUTE_32F_FAST_TF32",
                name=node.parent.generate_node_name("GEMMFP8"),
            )
        )
        if do_reshape:
            # One of the inputs had 3 dimensions.
            index = 0 if self.was_reshaped[0] else 1
            fname = f"MatMulReshapeTransposeBack{index}"
            mmshape = node.parent.generate_name(f"{name}_mmshape")

            shape = self.was_reshaped[index]
            if isinstance(shape, tuple):
                self.added.append(
                    make_node(
                        "Constant",
                        [],
                        [mmshape],
                        value=make_tensor(
                            mmshape, TensorProto.INT64, [len(shape)], list(shape)
                        ),
                    )
                )
            elif isinstance(shape, str):
                self.added.append(make_node("Shape", [shape], [mmshape]))
            else:
                raise TypeError(
                    f"Unexpected shape={shape} in was_reshaped={self.was_reshaped}."
                )

            self.added.append(
                make_node(
                    fname,
                    [gemm_outputs[0], mmshape],
                    node.outputs,
                    perm=[1, 0],
                    domain=self.main_state.domain_dq,
                    name=node.parent.generate_node_name("MMRTB"),
                )
            )
            if (
                self.main_state.domain_dq,
                fname,
            ) not in self.main_state.local_functions:
                # use local functions
                self.main_state.local_functions[
                    self.main_state.domain_dq, fname
                ] = make_matmul_reshape_transpose_back_function_proto(
                    domain=self.main_state.domain_dq,
                    opset=self.main_state.opset,
                    index=index,
                )

            elif self.was_reshaped[1]:
                raise NotImplementedError(
                    f"Shape is {shape}. This case is not implemented yet."
                )

        self.removed.append(node)
        opsets = {self.main_state.domain_gemm: 1}
        if self.main_state.domain_dq != "":
            opsets.update({self.main_state.domain_dq: 1})
        return opsets


def quantize_float8_matmul(
    node: Node,
    elem_type: int,
    output_type: int,
    version: str,
    opset: int,
    local_functions: Optional[Dict[str, FunctionProto]] = None,
    index_transpose: int = 2,
    domain_ops: Optional[Dict[str, str]] = None,
    quantize_options: QuantizeOptions = QuantizeOptions.NONE,
) -> Optional[TransformResults]:
    """
    Quantizes matrix multiplications.

    :param node: matrix multiplication
    :param elem_type: float 8 type to quantize into
    :param output_type: output type, result of the quantization
    :param version: `'onnxruntime'` to use operators from onnx and onnxruntime,
        `'onnx-extended'` to use experimental operators
    :param opset: main opset (used to specify local functions)
    :param local_functions: None to avoid using local functions,
        otherwise a dictionary with the existing local functions to
        add to the model
    :param quiet: True to silently skip failing nodes
    :param index_transpose: which input to transpose before calling gemm:
        0 (none), 1 (first), 2 (second), 3 for both
    :param domain_ops: domain to use for operators used as keys in the dictionary
    :param quantize_options: see :class:`QuantizeOptions`
    :return: nodes to remove, nodes to add, new opsets

    **About tensors with more than two dimensions.**

    Gemm only supports matrices or 2D tensors.
    When one input does not follow that schema,
    it uses the following tricks. First one when the first input
    has more than one dimension. We could describe that
    with the Einstein summation. We want `abij, jk -> abik`.

    ::

        abij, jk -> abik
        abij ~> Cj
        Cj, jk -> Ck      # gemm
        Ck ~> abik

    In python:

    .. runpython::
        :showcode:

        import numpy as np

        a = np.arange(16).reshape((2, 2, 2, 2))
        b = np.arange(4).reshape((2, 2)) * 10

        expected = a @ b

        a2 = a.reshape((-1, a.shape[-1]))
        b2 = b
        res = a2 @ b2
        final = res.reshape(a.shape[:-2] + (-1, res.shape[-1]))

        print("------")
        print(a)
        print(b)
        print("------")
        print(expected)
        print("------")
        print(final)
        assert expected.tolist() == final.tolist()

    For the other, we use the trick `a @ b = (b' @ a')'`.
    Second one when the second input has more than one dimension.
    We want `ij, abjk -> abik`.

    ::

        ij, abjk -> abik
        abjk ~> Cjk ~> jCk ~> jD
        ij, jD -> iD      # gemm
        iD ~> iCk ~> Cik ~> abik

    .. runpython::
        :showcode:

        import numpy as np

        a = np.arange(4).reshape((2, 2)) * 10
        b = np.arange(16).reshape((2, 2, 2, 2))

        expected = a @ b

        a2 = a
        b2 = (
            b.reshape((-1,) + b.shape[-2:])
            .transpose((1, 0, 2))
            .reshape((b.shape[-2], -1))
        )
        res = a2 @ b2
        final = (
            res.reshape(a.shape[0], -1, b.shape[-1])
            .transpose((1, 0, 2))
            .reshape(b.shape[:-2] + (-1, b.shape[-1]))
        )

        print("------")
        print(a)
        print(b)
        print("------")
        print(expected)
        print("------")
        print(final)
        assert expected.tolist() == final.tolist()

    Both sides cannot have more than two dimensions in the current implementation.
    """
    main_state = _MainQuantizeState(
        version, domain_ops, local_functions, opset, index_transpose
    )

    if node.op_type == "MatMul":
        quantize_state = _QuantizeState(logger, quantize_options, main_state)
        for index, name in enumerate(node.inputs):
            do_transpose = bool((1 << index) & index_transpose)
            if node.parent.is_constant(name):
                # Quantized constant weights
                quantize_state.quantize_connstant_weights(
                    node, name, elem_type, index, do_transpose
                )
            else:
                quantize_state.quantize_dynamic(
                    node, name, elem_type, index, do_transpose
                )

        opsets = quantize_state.finalize(node, name, output_type)

        return TransformResults(
            removed_nodes=quantize_state.removed,
            added_nodes=quantize_state.added,
            new_opsets=opsets,
            local_functions=main_state.local_functions,
        )

    raise NotImplementedError(
        f"Quantization into float 8 not yet implemented for {node.op_type!r}."
    )


def quantize_float8(
    graph: Graph,
    elem_type: int = TensorProto.FLOAT8E4M3FN,
    output_type: int = TensorProto.FLOAT,
    early_stop: int = -1,
    version: str = "onnxruntime",
    quiet: bool = False,
    index_transpose: int = 2,
    domain_ops: Optional[Dict[str, str]] = None,
    exceptions: Optional[List[Dict[str, str]]] = None,
    quantize_options: QuantizeOptions = QuantizeOptions.NONE,
) -> Optional[Graph]:
    """
    Transforms a graph to introduce quantized weights.
    This transformation requires opset 20. The graph is
    upgraded if the main opset is below. It is better to do
    it before calling this function.

    :param graph: Graph
    :param elem_type: quantization type
    :param output_type: output type
    :param early_stop: -1 to go through all nodes or a value `n > 0`
        to stop after n changes
    :param version: `'onnxruntime'` to use operators from onnx and onnxruntime,
        `'onnx-extended'` to use experimental operators
    :param quiet: catch exception and silently skip failing nodes
    :param index_transpose: which input to transpose before calling gemm:
        0 (none), 1 (first), 2 (second), 3 for both
    :param domain_ops: domain to use for operators used as keys in the dictionary
    :param exceptions: exclude nodes from the quantization,
        `[{"name": "node_name1"}, {"name": "node_name2"}]` will exclude
        these two node names from the quantization
    :param quantize_options: see :class:`QuantizeOptions`
    :return: Graph or None if not modified

    Transformation are logged with logger `onnx-extended/transformer`.
    The graph is modified inplace.
    Enables the logs gives a better idea of the progress.
    """
    main_opset = graph.get_opset("")
    if main_opset < 19:
        logger.info(
            "[quantize_float8] upgrade model from opset %d to %s", main_opset, 19
        )
        graph.upgrade_opsets({"": 19})
        main_opset = 19

    if len(graph.functions) > 0:
        raise NotImplementedError("Quantization of local functions is not implemented.")
    local_functions = graph.functions.copy()
    n_local_functions = len(local_functions)

    new_opsets = {}
    to_add = []
    n_nodes = len(graph)
    n_changes = 0
    for index, node in enumerate(graph):
        if exceptions is not None:
            # checks if a node is not excluded from the list to convert
            has_matched = False
            for exc in exceptions:
                if node.match(exc):
                    has_matched = True
                    break
            if has_matched:
                continue

        if node.op_type in {"MatMul", "Gemm"}:
            logger.info("[quantize_float8] %d/%d quantize %s", index, n_nodes, node)
            try:
                results = quantize_float8_matmul(
                    node,
                    elem_type,
                    output_type,
                    version=version,
                    opset=main_opset,
                    local_functions=local_functions,
                    index_transpose=index_transpose,
                    domain_ops=domain_ops,
                    quantize_options=quantize_options,
                )
            except (QuantizationError, NotImplementedError) as e:
                if quiet:
                    logger.warn(
                        "[quantize_float8] %d/%d failed to quantize due to %s",
                        index,
                        n_nodes,
                        e,
                    )
                    continue
                raise e

            if len(local_functions) != len(results.local_functions) or id(
                local_functions
            ) != id(results.local_functions):
                raise RuntimeError(
                    f"Inconsistency lenghts: "
                    f"{len(local_functions)} != {len(results.local_functions)}, "
                    f"ids: {id(local_functions)} != {id(results.local_functions)}."
                )

            if results is None:
                continue
            rem, add = results.removed_nodes, results.added_nodes
            to_add.append((rem, add))
            if len(results.new_opsets) > 0:
                n_changes += 1
                new_opsets.update(results.new_opsets)
            if early_stop > 0 and n_changes >= early_stop:
                break

    if len(to_add) == 0:
        return None

    for rem, add in to_add:
        for r in rem:
            logger.debug("[quantize_float8] del %s", r)
        added = graph.replace_nodes([r.index for r in rem], add, new_opsets)
        for a in added:
            logger.debug("[quantize_float8] add %s", a)

    graph.simplify()
    if local_functions is not None and len(local_functions) > n_local_functions:
        graph.add_functions(local_functions.values())
    return graph


def _cast_constant(
    node: Node,
    from_type: int,
    to_type: int,
) -> Optional[TransformResults]:
    """
    Converts a node if it is a tensor of element type *from_type*
    and cast it into *to_type*.

    :param node: node
    :param from_type: element type to replace
    :param to_type: type to cast into
    :return: TransformResults
    """
    op_type = node.op_type

    if op_type in {"Constant", "initializer"}:
        cst = node.get_tensor()
        if cst.data_type != from_type:
            return None

        arr = to_array_extended(cst)
        try:
            cast = Cast.eval(arr, to=to_type)
        except SchemaError:
            # cast does not work
            np_type = tensor_dtype_to_np_dtype(to_type)
            cast = arr.astype(np_type)
        if op_type == "initializer":
            return TransformResults(
                removed_nodes=[node],
                added_nodes=[from_array_extended(cast, name=node.proto.name)],
            )
        if op_type == "Constant":
            return TransformResults(
                removed_nodes=[node],
                added_nodes=[
                    make_node(
                        "Constant",
                        [],
                        node.outputs,
                        value=from_array_extended(cast, name=node.outputs[0]),
                    )
                ],
            )
    if op_type in {"input", "output"}:
        if isinstance(node.proto, str):
            # A FunctionProto input, nothing to do.
            return None
        ttype = node.proto.type
        if not ttype.tensor_type or ttype.tensor_type.elem_type != from_type:
            return None
        return TransformResults(
            removed_nodes=[node],
            added_nodes=[
                make_tensor_value_info(
                    node.proto.name,
                    to_type,
                    shape=_get_shape(ttype),
                    doc_string=node.proto.doc_string,
                    shape_denotation=ttype.denotation,
                )
            ],
        )
    if op_type in {"Cast"}:
        to = node.getattr("to", int)
        if to != from_type:
            return None
        return TransformResults(
            removed_nodes=[node],
            added_nodes=[make_node("Cast", node.inputs, node.outputs, to=to_type)],
        )

    raise RuntimeError(f"Unexpected node type {op_type!r}.")


def cast_constant(
    graph: Graph,
    from_type: int = TensorProto.FLOAT,
    to_type: int = TensorProto.FLOAT16,
    quiet: bool = False,
) -> Optional[Graph]:
    """
    Converts all constants and initializers to the same type.
    It also modifies the input.

    :param graph: Graph
    :param from_type: type of the constants to convert
    :param to_type: new type for the constants
    :param quiet: catch exception and silently skip failing nodes
    :return: Graph or None if not modified

    Transformation are logged with logger `onnx-extended/transformer`.
    The graph is modified inplace.
    Enables the logs gives a better idea of the progress.
    """
    if len(graph.functions) > 0:
        raise NotImplementedError("Conversion of local functions is not implemented.")

    to_add = []
    n_nodes = len(graph)
    for index, node in enumerate(graph):
        if node.op_type in {"Constant", "initializer", "input", "output", "Cast"}:
            logger.info("[cast_constant] %d/%d convert %s", index, n_nodes, node)
            results = _cast_constant(node, from_type, to_type)
            if results is None:
                continue
            rem, add = results.removed_nodes, results.added_nodes
            to_add.append((rem, add))

    if len(to_add) == 0:
        return None

    for rem, add in to_add:
        for r in rem:
            logger.debug("[cast_constant] del %s", r)
        added = graph.replace_nodes([r.index for r in rem], add)
        for a in added:
            logger.debug("[cast_constant] add %s", a)

    return graph.simplify()



import json
import sys
from textwrap import dedent
from argparse import ArgumentParser
from onnx_extended.tools.run_onnx import TestRun


def get_parser():
    parser = ArgumentParser(
        prog="run_onnx_main",
        description=dedent(
            """
        Runs a Benchmark.
        """
        ),
    )
    parser.add_argument(
        "-p", "--path", type=str, required=True, help="path to the test definition"
    )
    parser.add_argument("-i", "--index", type=int, default=0)
    parser.add_argument("-r", "--repeat", type=int, default=10)
    parser.add_argument("-w", "--warmup", type=int, default=5)
    parser.add_argument(
        "-e",
        "--runtime",
        default="ReferenceEvaluator",
        help="""
        A choice among
        CReferenceEvaluator, ReferenceEvaluator, onnxruntime, CustomTreeEnsemble.x.x.x,
        CustomTreeEnsemble.x.x.x is based on a custom node,
        with different settings.
        """,
    )
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    tr = TestRun(args.path)
    if args.runtime == "CReferenceEvaluator":
        from onnx_extended.reference import CReferenceEvaluator

        f_build = lambda proto: CReferenceEvaluator(proto)
        f_run = lambda rt, feeds: rt.run(None, feeds)
    elif args.runtime == "ReferenceEvaluator":
        from onnx.reference import ReferenceEvaluator

        f_build = lambda proto: ReferenceEvaluator(proto)
        f_run = lambda rt, feeds: rt.run(None, feeds)
    elif args.runtime == "onnxruntime":
        from onnxruntime import InferenceSession

        f_build = lambda proto: InferenceSession(
            proto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        f_run = lambda rt, feeds: rt.run(None, feeds)
    elif args.runtime.startswith("CustomTreeEnsemble"):
        from onnx_extended.ortops.optim.cpu import get_ort_ext_libs
        from onnx_extended.ortops.optim.optimize import (
            change_onnx_operator_domain,
            get_node_attribute,
        )
        from onnxruntime import InferenceSession, SessionOptions

        spl = args.runtime.split(".")
        op_name = spl[0].replace("Custom", "")
        params = [int(a) for a in spl[1:]]
        if len(params) > 6:
            raise ValueError(f"Unexpected runtime {args.runtime!r}.")

        optim_params = {}
        for i, k in enumerate(
            [
                "parallel_tree",
                "parallel_tree_N",
                "parallel_N",
                "batch_size_tree",
                "batch_size_rows",
                "use_node3",
            ]
        ):
            if i >= len(params):
                break
            optim_params[k] = params[i]

        def transform_model(onx, **kwargs):
            att = get_node_attribute(onx.graph.node[0], "nodes_modes")
            modes = ",".join(map(lambda s: s.decode("ascii"), att.strings))
            return change_onnx_operator_domain(
                onx,
                op_type=op_name,
                op_domain="ai.onnx.ml",
                new_op_domain="onnx_extented.ortops.optim.cpu",
                nodes_modes=modes,
                **kwargs,
            )

        opts = SessionOptions()
        r = get_ort_ext_libs()
        opts.register_custom_ops_library(r[0])

        f_build = lambda proto, opts=opts: InferenceSession(
            transform_model(proto, **optim_params).SerializeToString(),
            opts,
            providers=["CPUExecutionProvider"],
        )
        f_run = lambda rt, feeds: rt.run(None, feeds)

    else:
        raise ValueError(f"Unexpected value {args.runtime!r} for runtime.")

    check = tr.test(
        exc=False,
        f_build=f_build,
        f_run=f_run,
        index=args.index,
    )
    bench = tr.bench(
        f_build=f_build,
        f_run=f_run,
        warmup=args.warmup,
        repeat=args.repeat,
        index=args.index,
    )
    output = {"test": check, "bench": bench, "runtime": args.runtime}
    js = json.dumps(output)
    print(js)


if __name__ == "__main__":
    main(sys.argv[1:])



import time
from itertools import product
from typing import Any, Callable, Dict, List, Optional, Union
import numpy
from onnx import AttributeProto, ModelProto, NodeProto, GraphProto, FunctionProto
from onnx.helper import make_model, make_node, make_graph, make_opsetid
from ...ext_test_case import measure_time


def has_subgraph(node: NodeProto) -> bool:
    """
    Tells if a node has a subgraph as an attribute.
    """
    for att in node.attribute:
        if att.type == AttributeProto.GRAPH:
            return True
    return False


def get_node_attribute(node: NodeProto, name: str) -> AttributeProto:
    """
    Returns the value of one attribute.

    :param node: node
    :param name: attribute name
    :return: value
    """
    for att in node.attribute:
        if att.name == name:
            return att
    raise KeyError(
        f"Unable to find {name!r} among {list(att.name for att in node.attribute)}."
    )


def change_onnx_operator_domain(
    onx: Union[ModelProto, GraphProto, FunctionProto],
    op_type: str,
    op_domain: str = "",
    new_op_type: Optional[str] = None,
    new_op_domain: Optional[str] = None,
    new_opset: Optional[int] = None,
    **kwargs: Dict[str, Any],
) -> Union[ModelProto, GraphProto, FunctionProto]:
    """
    Replaces an operator by another one in the same domain
    or another one.

    :param onx: proto to modify
    :param op_type: operator to look for
    :param op_domain: domain to look for
    :param new_op_type: new operator name or None for the same name
    :param new_op_domain: new domain name or None the for the same domain
    :param new_opset: new opset for the new domain
    :param kwargs: modified parameters, set it to None to remove them
    :return: same type as the input

    The function is not recursive yet.
    """

    def change_node(node):
        atts = []
        new_kwargs = {}
        for att in node.attribute:
            if att.name in kwargs:
                v = kwargs[att.name]
                if v is None:
                    continue
                new_kwargs[att.name] = v
                continue
            atts.append(att)
        for k, v in kwargs.items():
            if v is None or k in new_kwargs:
                continue
            new_kwargs[k] = v
        new_node = make_node(
            new_op_type or node.op_type,
            node.input,
            node.output,
            domain=new_op_domain or node.domain,
            **new_kwargs,
        )
        if len(atts) > 0:
            new_node.attribute.extend(atts)
        return new_node

    if isinstance(onx, GraphProto):
        new_nodes = []
        modified = False
        for node in onx.node:
            if has_subgraph(node):
                raise NotImplementedError(
                    f"The function is not recursive yet and cannot "
                    f"handle node {node.op_type!r} from domain "
                    f"{node.domain!r}."
                )
            if node.op_type == op_type and node.domain == op_domain:
                new_node = change_node(node)
                new_nodes.append(new_node)
                modified = True
                continue
            new_nodes.append(node)
        if not modified:
            return onx
        return make_graph(
            new_nodes,
            onx.name,
            onx.input,
            onx.output,
            onx.initializer,
            onx.sparse_initializer,
        )

    if isinstance(onx, FunctionProto):
        raise NotImplementedError()

    if not isinstance(onx, ModelProto):
        raise TypeError(f"Unexpected type for onx {type(onx)}.")

    new_graph = change_onnx_operator_domain(
        onx.graph,
        op_type=op_type,
        op_domain=op_domain,
        new_opset=new_opset,
        new_op_type=new_op_type,
        new_op_domain=new_op_domain,
        **kwargs,
    )
    if id(new_graph) == id(onx.graph):
        # no change
        return onx

    if new_op_domain is None:
        new_op_domain = op_domain
    if new_op_domain == op_domain and new_opset is not None:
        raise ValueError(
            f"If new_op_domain==domain=={new_op_domain!r}, "
            f"new_opset must be None not {new_opset}."
        )
    opsets = list(onx.opset_import)
    if new_op_domain != op_domain:
        opsets.append(make_opsetid(new_op_domain, new_opset or 1))

    new_model = make_model(
        new_graph,
        functions=onx.functions,
        ir_version=onx.ir_version,
        producer_name=onx.producer_name,
        producer_version=onx.producer_version,
        model_version=onx.model_version,
        doc_string=onx.doc_string,
        opset_imports=opsets,
        domain=onx.domain,
    )
    return new_model


def optimize_model(
    onx: ModelProto,
    feeds: Dict[str, numpy.ndarray],
    transform: Callable[[ModelProto], ModelProto],
    session: Callable[[ModelProto], Any],
    params: Dict[str, List[Any]],
    baseline: Optional[Callable[[ModelProto], Any]] = None,
    verbose: bool = False,
    number: int = 10,
    repeat: int = 10,
    warmup: int = 5,
    n_tries: int = 2,
    sleep: float = 0.1,
) -> List[Dict[str, Union[str, float]]]:
    """
    Optimizes a model by trying out many possibilities.

    :param onx: ModelProto
    :param feeds: inputs as a dictionary of numpy arrays
    :param transform: function taking a ModelProto and returning a ModelProto
        based on the values coming from *params*
    :param session: function which takes a modifed ModelProto
        and return a session
    :param params: dictionary of values to test `{ param_name: [ param_values ] }`
    :param baseline: function which takes a modifed ModelProto
        and return a session, identified as the baseline
    :param verbose: use :epkg:`tqdm` to show improvment
    :param number: parameter to :func:`measure_time
        <onnx_extended.ext_test_case.measure_time>`
    :param repeat: parameter to :func:`measure_time
        <onnx_extended.ext_test_case.measure_time>`
    :param warmup: parameter to :func:`measure_time
        <onnx_extended.ext_test_case.measure_time>`
    :param n_tries: number of times to measure, if the measurements returns
        very different results, values for *number* or *repeat* should
        be increased
    :param sleep: time to sleep between two measurements
    :return: list of results returned by :func:`measure_time
        <onnx_extended.ext_test_case.measure_time>`

    See example :ref:`l-plot-optim-tree-ensemble` for an example.
    """
    if sleep >= 1:
        raise ValueError(f"sleep={sleep} >= 1, probably a mistake.")
    keys = ["TRY"] + list(params.keys())
    sets = [list(range(n_tries))] + [params[k] for k in keys[1:]]
    loops = list(product(*sets))
    if verbose:
        from tqdm import tqdm

        loop = tqdm(loops)
    else:
        loop = loops

    res = []
    if baseline is not None:
        sess = baseline(onx)
        # one run to make run it is working
        sess.run(None, feeds)
        if sleep > 0:
            time.sleep(sleep)
        obs = measure_time(
            lambda sess=sess: sess.run(None, feeds),
            number=number,
            repeat=repeat,
            warmup=warmup,
        )
        obs["n_exp"] = 0
        obs["n_exp_name"] = "TRY=0,baseline"
        obs["short_name"] = "0,baseline"
        obs["TRY"] = 0
        obs["name"] = "baseline"
        res.append(obs)

    for it, values in enumerate(loop):
        if verbose:
            msg = [f"i={it+1}/{len(loops)}"]
            msg.extend([f"{k}={v}" for k, v in zip(keys, values)])
            loop.set_description(" ".join(msg))

        kwargs = dict(zip(keys, values))
        del kwargs["TRY"]
        onx_modified = transform(onx, **kwargs)
        sess = session(onx_modified)
        if sleep > 0:
            time.sleep(sleep)
        obs = measure_time(
            lambda sess=sess: sess.run(None, feeds),
            number=number,
            repeat=repeat,
            warmup=warmup,
        )
        obs.update(kwargs)
        obs["n_exp"] = it
        obs["n_exp_name"] = ",".join(f"{k}={v}" for k, v in zip(keys, values))
        obs["short_name"] = ",".join(f"{v}" for v in values)
        obs["name"] = ",".join(f"{v}" for v in values[1:])
        res.append(obs)

    if baseline is not None:
        for n in range(1, n_tries):
            sess = baseline(onx)
            if sleep > 0:
                time.sleep(sleep)
            obs = measure_time(
                lambda sess=sess: sess.run(None, feeds),
                number=number,
                repeat=repeat,
                warmup=warmup,
            )
            obs["n_exp"] = 0
            obs["n_exp_name"] = f"TRY={n},baseline"
            obs["short_name"] = f"{n},baseline"
            obs["name"] = "baseline"
            obs["TRY"] = n
            res.append(obs)

    return res



from onnx import FunctionProto, TensorProto
from onnx.helper import (
    make_function,
    make_node,
    make_opsetid,
    make_tensor,
)


def make_matmul_reshape_transpose_function_proto(
    domain: str, opset: int, index: int, transpose: bool
) -> FunctionProto:
    """
    Creates the FunctionProto for function
    `MatMulReshapeTranspose[index]`
    to reshape in two dimensions and transpose an input
    for operator Matmul. The transposition is optional.

    :param domain: local domain name
    :param opset: opset to use to define the function
    :param index: which input it is, 0 for left, 1 for right
    :param transpose: transpose the input
    :return: FunctionProto

    The function takes 1 input and returns 1 output.
    See :func:`quantize_float8_matmul`.
    """

    if index == 0:
        nodes = [
            make_node("Shape", ["x"], ["shape_x"]),
            make_node(
                "Constant",
                [],
                ["m1"],
                value=make_tensor("m1", TensorProto.INT64, [1], [-1]),
            ),
            make_node("Gather", ["shape_x", "m1"], ["last_dim"]),
            make_node("Concat", ["m1", "last_dim"], ["new_shape"], axis=0),
            make_node("Reshape", ["x", "new_shape"], ["reshaped_name"]),
        ]
    elif index == 1:
        # values = (
        #     values.reshape((-1,) + values.shape[-2:])
        #     .transpose((1, 0, 2))
        #     .reshape((values.shape[-2], -1))
        # )
        nodes = [
            make_node("Shape", ["x"], ["shape_x"]),
            make_node(
                "Constant",
                [],
                ["m1"],
                value=make_tensor("m1", TensorProto.INT64, [1], [-1]),
            ),
            make_node(
                "Constant",
                [],
                ["m2"],
                value=make_tensor("m2", TensorProto.INT64, [1], [-2]),
            ),
            make_node(
                "Constant",
                [],
                ["m21"],
                value=make_tensor("m2", TensorProto.INT64, [2], [-2, -1]),
            ),
            make_node("Gather", ["shape_x", "m21"], ["last_dims"]),
            make_node("Concat", ["m1", "last_dims"], ["new_shape"], axis=0),
            make_node("Reshape", ["x", "new_shape"], ["xsh"]),
            make_node("Transpose", ["xsh"], ["xtr"], perm=[1, 0, 2]),
            make_node("Gather", ["shape_x", "m2"], ["bldims"]),
            make_node("Concat", ["bldims", "m1"], ["new_shape2"], axis=0),
            make_node("Reshape", ["xtr", "new_shape2"], ["reshaped_name"]),
        ]
    else:
        raise ValueError(f"index must be 0 or 1 not {index}.")

    if transpose:
        nodes.append(make_node("Transpose", ["reshaped_name"], ["y"], perm=[1, 0]))
    else:
        nodes.append(make_node("Identity", ["reshaped_name"], ["y"]))

    return make_function(
        domain,
        f"MatMulReshapeTranspose{'T' if transpose else 'N'}{index}",
        ["x"],
        ["y"],
        nodes,
        opset_imports=[make_opsetid("", opset)],
    )


def make_matmul_reshape_transpose_back_function_proto(
    domain: str, opset: int, index: int
) -> FunctionProto:
    """
    Creates the FunctionProto for function `ReshapeTransposeBack[index]`
    to reshape with more two dimensions an input which was modified
    by `ReshapeTransposeBack[index]`

    :param domain: local domain name
    :param opset: opset to use to define the function
    :param index: which input it is, 0 for left, 1 for right
    :return: FunctionProto

    The function takes 2 inputs `(x, shape)` and returns 1 output.
    See :func:`quantize_float8_matmul`.
    """

    if index == 0:
        # res.reshape(a.shape[:-2] + (-1, res.shape[-1]))
        nodes = [
            make_node(
                "Constant",
                [],
                ["zero"],
                value=make_tensor("zero", TensorProto.INT64, [1], [0]),
            ),
            make_node(
                "Constant",
                [],
                ["m1"],
                value=make_tensor("m1", TensorProto.INT64, [1], [-1]),
            ),
            make_node(
                "Constant",
                [],
                ["m2"],
                value=make_tensor("m2", TensorProto.INT64, [1], [-2]),
            ),
            make_node("Slice", ["shape", "zero", "m2", "zero"], ["sliced"]),
            make_node("Gather", ["shape", "m2"], ["shm2"]),
            make_node("Concat", ["sliced", "shm2", "m1"], ["new_shape"], axis=0),
            make_node("Reshape", ["x", "new_shape"], ["y"]),
        ]
    elif index == 1:
        # final = (
        #     res.reshape(a.shape[0], -1, b.shape[-1])
        #     .transpose((1, 0, 2))
        #     .reshape(b.shape[:-2] + (-1, b.shape[-1]))
        # )
        nodes = [
            make_node(
                "Constant",
                [],
                ["zero"],
                value=make_tensor("zero", TensorProto.INT64, [1], [0]),
            ),
            make_node(
                "Constant",
                [],
                ["m1"],
                value=make_tensor("m1", TensorProto.INT64, [1], [-1]),
            ),
            make_node(
                "Constant",
                [],
                ["m2"],
                value=make_tensor("m2", TensorProto.INT64, [1], [-2]),
            ),
            make_node("Shape", ["x"], ["shape_x"]),
            make_node("Gather", ["shape_x", "zero"], ["d0"]),
            make_node("Gather", ["shape", "m1"], ["last_dim"]),
            make_node("Concat", ["d0", "m1", "last_dim"], ["new_shape"], axis=0),
            make_node("Reshape", ["x", "new_shape"], ["xsh"]),
            make_node("Transpose", ["xsh"], ["xtr"], perm=[1, 0, 2]),
            make_node("Slice", ["shape", "zero", "m2", "zero"], ["sliced"]),
            make_node("Concat", ["sliced", "m1", "last_dim"], ["final_shape"], axis=0),
            make_node("Reshape", ["xtr", "final_shape"], ["y"]),
        ]

    else:
        raise ValueError(f"index must be 0 or 1 not {index}.")

    return make_function(
        domain,
        f"MatMulReshapeTransposeBack{index}",
        ["x", "shape"],
        ["y"],
        nodes,
        opset_imports=[make_opsetid("", opset)],
    )



def cuda_version() -> str:
    """
    Returns the cuda version it was compiled with.
    If CUDA was not available, it retunrs `"0.0"`.
    """
    try:
        from .cuda_example_py import cuda_version as cv
    except ImportError:
        # No CUDA
        return "0.0"
    v = cv()
    major = v // 1000
    minor = (v % 1000) // 10
    return f"{major}.{minor}"



import os
from typing import Any, Dict, List, Optional, Set, Union
import numpy as np
from onnx import FunctionProto, ModelProto, NodeProto, TensorProto
from onnx.defs import get_schema
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array
from onnx.reference import ReferenceEvaluator
from onnx.reference.custom_element_types import (
    bfloat16,
    float8e4m3fn,
    float8e4m3fnuz,
    float8e5m2,
    float8e5m2fnuz,
)
from onnx.reference.op_run import OpRun


def from_array_extended(tensor: np.ndarray, name: Optional[str] = None) -> TensorProto:
    """
    Converts an array into a TensorProto including float 8 types.

    :param tensor: numpy array
    :param name: name
    :return: TensorProto
    """
    dt = tensor.dtype
    if dt == float8e4m3fn and dt.descr[0][0] == "e4m3fn":
        to = TensorProto.FLOAT8E4M3FN
        dt_to = np.uint8  # type: ignore[assignment]
    elif dt == float8e4m3fnuz and dt.descr[0][0] == "e4m3fnuz":
        to = TensorProto.FLOAT8E4M3FNUZ
        dt_to = np.uint8  # type: ignore[assignment]
    elif dt == float8e5m2 and dt.descr[0][0] == "e5m2":
        to = TensorProto.FLOAT8E5M2
        dt_to = np.uint8  # type: ignore[assignment]
    elif dt == float8e5m2fnuz and dt.descr[0][0] == "e5m2fnuz":
        to = TensorProto.FLOAT8E5M2FNUZ
        dt_to = np.uint8  # type: ignore[assignment]
    elif dt == bfloat16 and dt.descr[0][0] == "bfloat16":
        to = TensorProto.BFLOAT16
        dt_to = np.uint16  # type: ignore[assignment]
    else:
        return from_array(tensor, name)

    t = from_array(tensor.astype(dt_to), name)
    t.data_type = to
    return t


class CReferenceEvaluator(ReferenceEvaluator):
    """
    This class replaces the python implementation by C implementation
    for a short list of operators quite slow in python (such as `Conv`).
    The class automatically replaces a python implementation
    by a C implementation if available. See example :ref:`l-example-conv`.

    ::

        from onnx.reference import ReferenceEvaluator
        from onnx_extended.reference.c_ops import Conv
        ref = ReferenceEvaluator(..., new_ops=[Conv])

    See :class:`onnx.reference.ReferenceEvaluator` for a detailed documentation.

    **Additions**

    Parameter **save_intermediate** can be set to a folder to save intermediate
    results in this folder. It follows the same design as the backend test.
    Let's consider a model with the following nodes:

    ::

        <
            ir_version: 8,
            opset_import: [ "" : 18]
        >
        agraph (float[N, 128] X, float[128,10] W, float[10] B) => (float[N] C)
        {
            T = MatMul(X, W)
            S = Add(T, B)
            C = Softmax(S)
        }

    It will produce the following files after it is run with
    `CReferenceEvaluator(..., save_intermediate="modelrun")`.

    ::

        modelrun
            +-- test_node_0_MatMul
            |       +-- model.onnx
            |       +-- test_data_set_0
            |               + input_0.pb
            |               + input_1.pb
            |               + output_0.pb
            +-- test_node_1_Add
            |       +-- model.onnx
            |       +-- test_data_set_0
            |               + input_0.pb
            |               + input_1.pb
            |               + output_0.pb
            +-- test_node_2_Softmax
                    +-- model.onnx
                    +-- test_data_set_0
                            + input_0.pb
                            + output_0.pb

    These files can then be run with a different runtime to look for discrepancies.
    Following example executes node by node with onnxruntime.

    ::

        from onnx.backend.test.loader import load_model_tests
        from onnx.reference.c_reference_backend import (
            ReferenceEvaluatorBackend,
            create_reference_backend,
        )
        from onnxruntime import InferenceSession

        root = "folder which folder modelrun"
        examples = load_model_tests(root, "modelrun")

        class Wrapper(InferenceSession):

            def __init__(self, model, *args, providers=None, **kwargs):
                super().__init__(
                    model.SerializeToString(),
                    *args,
                    providers=providers or ["CPUExecutionProvider"],
                    **kwargs,
                )

            def run(self, *args, **kwargs):
                return InferenceSession.run(self, *args, **kwargs)

            @property
            def input_names(self):
                return [i.name for i in self.get_inputs()]

            @property
            def output_names(self):
                return [o.name for o in self.get_outputs()]

        new_cls = ReferenceEvaluatorBackend[NewRef]
        backend = create_reference_backend(new_cls, path_to_test=root)
        beckend.run()

    .. versionadded:: 0.2.0
    """

    @staticmethod
    def default_ops():
        from onnx_extended.reference.c_ops.c_op_conv import Conv
        from onnx_extended.reference.c_ops.c_op_tree_ensemble_regressor import (
            TreeEnsembleRegressor_1,
            TreeEnsembleRegressor_3,
        )
        from onnx_extended.reference.c_ops.c_op_tree_ensemble_classifier import (
            TreeEnsembleClassifier_1,
            TreeEnsembleClassifier_3,
        )
        from onnx_extended.reference.c_custom_ops.custom_op_tree_ensemble_regressor import (  # noqa: E501
            TreeEnsembleRegressor_1 as TreeEnsembleRegressor_1_Float,
        )

        return [
            Conv,
            TreeEnsembleClassifier_1,
            TreeEnsembleClassifier_3,
            TreeEnsembleRegressor_1,
            TreeEnsembleRegressor_3,
            TreeEnsembleRegressor_1_Float,
        ]

    @staticmethod
    def filter_ops(
        proto: Union[FunctionProto, ModelProto],
        new_ops: List[OpRun],
        opsets: Dict[str, int],
    ) -> List[OpRun]:
        if opsets is None and isinstance(proto, (ModelProto, FunctionProto)):
            opsets = {d.domain: d.version for d in proto.opset_import}
        best = {}
        renamed = {}
        for cl in new_ops:
            if "_" not in cl.__name__:
                continue
            vers = cl.__name__.split("_")
            try:
                v = int(vers[-1])
            except ValueError:
                # not a version
                continue
            if opsets is not None and v > opsets.get(cl.op_domain, 1):
                continue
            renamed[cl.__name__] = cl
            key = cl.op_domain, "_".join(vers[:-1])
            if key not in best or best[key][0] < v:
                best[key] = (v, cl)

        modified = []
        for cl in new_ops:
            if cl.__name__ not in renamed:
                modified.append(cl)
        for k, v in best.items():
            atts = {"domain": k[0]}
            bases = (v[1],)
            if not hasattr(v[1], "op_schema"):
                atts["op_schema"] = get_schema(k[1], v[0], domain=v[1].op_domain)
            new_cl = type(k[1], bases, atts)
            modified.append(new_cl)

        new_ops = modified
        return new_ops

    def __init__(
        self,
        proto: Any,
        opsets: Optional[Dict[str, int]] = None,
        functions: Optional[List[Union[ReferenceEvaluator, FunctionProto]]] = None,
        verbose: int = 0,
        new_ops: Optional[List[OpRun]] = None,
        save_intermediate: Optional[str] = None,
        **kwargs,
    ):
        if new_ops is None:
            new_ops = CReferenceEvaluator.default_ops()
        else:
            new_ops = new_ops.copy()
            new_ops.extend(CReferenceEvaluator.default_ops())
        new_ops = CReferenceEvaluator.filter_ops(proto, new_ops, opsets)

        ReferenceEvaluator.__init__(
            self,
            proto,
            opsets=opsets,
            functions=functions,
            verbose=verbose,
            new_ops=new_ops,
            **kwargs,
        )

        self.save_intermediate = save_intermediate
        if save_intermediate is not None:
            self._cached_saved_results = {}

    def run(
        self,
        output_names,
        feed_inputs: Dict[str, Any],
        attributes: Optional[Dict[str, Any]] = None,
    ):
        """
        Executes the onnx model.

        :param output_names: requested outputs by names,
            None for all
        :param feed_inputs: dictionary `{ input name: input value }`
        :param attributes: attributes value if the instance runs a FunctionProto
        :return: list of requested outputs
        """
        if output_names is None:
            output_names = self.output_names
        if isinstance(self.proto_, FunctionProto) and attributes is None:
            raise TypeError(f"Unexpected proto type {type(self.proto_)}.")

        # step 1: inputs and initializers
        results = {"": None}
        results.update(self.rt_inits_)
        results.update(feed_inputs)
        for k, v in self.rt_inits_.items():
            self._log(2, " +C %s: %s", k, v)
        for k, v in feed_inputs.items():
            self._log(2, " +I %s: %s", k, v)

        # step 2: execute nodes
        for index, node in enumerate(self.rt_nodes_):
            self._log(1, "%s(%s) -> %s", node.op_type, node.input, node.output)
            inputs = [results[i] for i in node.input]
            linked_attributes = {}
            if node.has_linked_attribute and attributes:
                linked_attributes["linked_attributes"] = attributes
            if node.need_context():
                outputs = node.run(*inputs, context=results, **linked_attributes)
            else:
                outputs = node.run(*inputs, **linked_attributes)
            for name, value in zip(node.output, outputs):
                if isinstance(value, tuple):
                    raise TypeError(
                        f"Unexected type {type(value)} for output {name!r}."
                    )
                self._log(2, " + %s: %s", name, value)
                results[name] = value
            if self.save_intermediate is not None:
                hidden_names = self._retrieve_hidden_inputs(node.onnx_node)
                set_inputs = set(node.input)
                hidden = {
                    k: v
                    for k, v in results.items()
                    if k in hidden_names and k not in set_inputs
                }
                self._save_intermerdiate_results(index, node, inputs, outputs, hidden)

        # return the results
        list_results: list[Any] = []
        for name in output_names:
            if name not in results:
                raise RuntimeError(
                    f"Unable to find output name {name!r} "
                    f"in {sorted(results)}, proto is\n{self.proto_}"
                )
            list_results.append(results[name])
        return list_results

    @staticmethod
    def _retrieve_input_names(nodes: List[NodeProto]) -> Set[str]:
        inputs = set()
        for node in nodes:
            inputs |= set(node.input)
            for att in node.attribute:
                if att.g:
                    inputs |= CReferenceEvaluator._retrieve_input_names(att.g.node)
        return inputs

    def _retrieve_hidden_inputs(self, node: NodeProto) -> Set[str]:
        names = set()
        for att in node.attribute:
            if att.g:
                names |= CReferenceEvaluator._retrieve_input_names(att.g.node)
        return names

    def _save_intermerdiate_results(
        self,
        index: int,
        node: NodeProto,
        inputs: List[np.ndarray],
        outputs: List[np.ndarray],
        hidden: Dict[str, np.ndarray],
    ) -> str:
        """
        Saves intermediate results into a folder with
        the same organization as the backend tests.

        :param index: index of the node in the graph
        :param node: node proto
        :param inputs: inputs of the node
        :param outputs: outputs of the node, supposedly the expected outputs
        :param hidden: hidden variables if the node has a subgraph
        :return: test folder
        """

        def get_shape(t):
            return list(t.dims)

        input_protos = [
            from_array_extended(inp, name) for name, inp in zip(node.input, inputs)
        ]
        output_protos = [
            from_array_extended(out, name) for name, out in zip(node.output, outputs)
        ]
        constants = []
        if hidden:
            constants.extend(
                [
                    make_node("Constant", [], [name], value=from_array_extended(value))
                    for name, value in hidden.items()
                ]
            )
        model = make_model(
            make_graph(
                [*constants, node.onnx_node],
                f"test_{node.op_type}",
                [
                    make_tensor_value_info(i.name, i.data_type, get_shape(i))
                    for i in input_protos
                ],
                [
                    make_tensor_value_info(o.name, o.data_type, get_shape(o))
                    for o in output_protos
                ],
            ),
            opset_imports=self.proto_.opset_import,
        )
        model_bytes = model.SerializeToString()
        if model_bytes not in self._cached_saved_results:
            sub = f"test_node_{index}_{node.op_type}"
            path = os.path.join(self.save_intermediate, sub)
            if not os.path.exists(path):
                os.makedirs(path)
            self._cached_saved_results[model_bytes] = [path, -1]
            model_file = os.path.join(path, "model.onnx")
            with open(model_file, "wb") as f:
                f.write(model_bytes)
        path, n_example = self._cached_saved_results[model_bytes]
        self._cached_saved_results[model_bytes][1] += 1
        n_example += 1
        sub_path = os.path.join(path, f"test_data_set_{n_example}")
        if not os.path.exists(sub_path):
            os.mkdir(sub_path)
        for i, t in enumerate(input_protos):
            with open(os.path.join(sub_path, f"input_{i}.pb"), "wb") as f:
                f.write(t.SerializeToString())
        for i, t in enumerate(output_protos):
            with open(os.path.join(sub_path, f"output_{i}.pb"), "wb") as f:
                f.write(t.SerializeToString())
        return path



import os
import textwrap
from typing import List
from ... import _get_ort_ext_libs


def get_ort_ext_libs() -> List[str]:
    """
    Returns the list of libraries implementing new simple
    :epkg:`onnxruntime` kernels implemented for the
    :epkg:`CPUExecutionProvider`.
    """
    return _get_ort_ext_libs(os.path.dirname(__file__))


def documentation() -> List[str]:
    """
    Returns a list of rst string documenting every implemented kernels
    in this subfolder.
    """
    return list(
        map(
            textwrap.dedent,
            [
                """
    onnx_extented.ortops.tutorial.cpu.DynamicQuantizeLinear
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements DynamicQuantizeLinear opset 20.

    **Provider**
    
    CPUExecutionProvider
    
    **Attributes**

    * to: quantized type

    **Inputs**
    
    * X (T1): tensor of type T

    **Outputs**

    * Y (T2): quantized X
    * scale (TS): scale
    * Y (T2): zero point

    **Constraints**

    * T1: float, float 16
    * TS: float
    * T2: int8, uint8, float8e4m3fn, float8e4m3fnuz, float8e5m2, float8e5m2fnuz
    """,
                """
    onnx_extented.ortops.tutorial.cpu.MyCustomOp
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It does the sum of two tensors.

    **Provider**
    
    CPUExecutionProvider
    
    **Inputs**
    
    * X (T): tensor of type T
    * Y (T): tensor of type T

    **Outputs**

    * Z (T): addition of X, Y

    **Constraints**

    * T: float
    """,
                """
    onnx_extented.ortops.tutorial.cpu.MyCustomOpWithAttributes
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It does the sum of two tensors + a constant equal to
    `cst = att_float + att_int64 + att_string[0] + att_tensot[0]`.

    **Provider**
    
    CPUExecutionProvider
    
    **Attributes**

    * att_float: a float
    * att_int64: an integer
    * att_tensor: a tensor of any type and shape
    * att_string: a string
    
    **Inputs**
    
    * X (T): tensor of type T
    * Y (T): tensor of type T

    **Outputs**

    * Z (T): addition of X, Y + cst

    **Constraints**

    * T: float
    """,
            ],
        )
    )



import os
import platform
from typing import Dict, List

_ort_ext_libs_pathes: Dict[str, List[str]] = {}


def _get_ort_ext_libs(path: str) -> List[str]:
    """
    Returns the list of libraries implementing new simple
    :epkg:`onnxruntime` kernels and places in folder *path*.
    """
    global _ort_ext_libs_pathes
    if path not in _ort_ext_libs_pathes:
        _ort_ext_libs_pathes[path] = []
    if len(_ort_ext_libs_pathes[path]) == 0:
        if platform.system() == "Windows":
            ext = ".dll"
        elif platform.system() == "Darwin":
            ext = ".dylib"
        else:
            ext = ".so"
        this = os.path.abspath(path)
        files = os.listdir(this)
        res = []
        for name in files:
            e = os.path.splitext(name)[-1]
            if e == ext and "ortops" in name:
                res.append(os.path.join(this, name))
        if len(res) == 0:
            raise RuntimeError(
                f"Unable to find any kernel library with ext={ext!r} "
                f"in {this!r} among {files}."
            )
        _ort_ext_libs_pathes[path] = res
    return _ort_ext_libs_pathes[path]



import os
from typing import Dict, Generator, Iterator, Optional, Set, Tuple, Union
import onnx

_rev_type: Dict[int, str] = {
    getattr(onnx.TensorProto, k): k
    for k in dir(onnx.TensorProto)
    if isinstance(getattr(onnx.TensorProto, k), int)
}


def load_model(
    model: Union[str, onnx.ModelProto, onnx.GraphProto, onnx.FunctionProto],
    external: bool = True,
    base_dir: Optional[str] = None,
) -> Union[onnx.ModelProto, onnx.GraphProto, onnx.FunctionProto]:
    """
    Loads a model or returns the only argument if the type
    is already a ModelProto.

    :param model: proto file
    :param external: loads the external data as well
    :param base_dir: needed if external is True and
        the model has external weights
    :return: ModelProto
    """
    if isinstance(model, onnx.ModelProto):
        if base_dir is not None and external:
            if not os.path.exists(base_dir):
                raise FileNotFoundError(f"Unable to find folder {base_dir!r}.")
            onnx.load_external_data_for_model(model, base_dir)
        return model
    if isinstance(model, (onnx.GraphProto, onnx.FunctionProto)):
        return model
    if not os.path.exists(model):
        raise FileNotFoundError(f"Unable to find model {model!r}.")
    with open(model, "rb") as f:
        return onnx.load(f, load_external_data=external)


def load_external(
    model: onnx.ModelProto, base_dir: str, names: Optional[Set[str]] = None
):
    """
    Loads external data into memory.

    :param model: the model loaded with :func:`load_model`
    :param base_dir: directory when the data can be found
    :param names: subsets of names to load or None for all
    """
    from onnx.external_data_helper import (
        _get_all_tensors,
        uses_external_data,
        load_external_data_for_tensor,
    )

    for tensor in _get_all_tensors(model):
        if names is not None and tensor.name not in names:
            continue
        if uses_external_data(tensor):
            load_external_data_for_tensor(tensor, base_dir)
            # After loading raw_data from external_data, change the state of tensors
            tensor.data_location = onnx.TensorProto.DEFAULT
            # and remove external data
            del tensor.external_data[:]


def enumerate_model_tensors(
    model: onnx.ModelProto,
) -> Iterator[Tuple[onnx.TensorProto, bool]]:
    """
    Enumerates all tensors in a model.

    :param model: model to process
    :return: iterator on a couple (TensorProto, bool),
        the boolean indicates if the data is external
    """
    from onnx.external_data_helper import (
        _get_all_tensors,
        uses_external_data,
    )

    for tensor in _get_all_tensors(model):
        yield tensor, uses_external_data(tensor)


def save_model(
    proto: onnx.ModelProto,
    filename: str,
    external: bool = False,
    convert_attribute: bool = True,
    size_threshold: int = 1024,
    all_tensors_to_one_file: bool = True,
):
    """
    Saves a model into an onnx file.

    :param proto: ModelProto
    :param filename: where to save it
    :param external: saves weights as external data
    :param convert_attribute: converts attributes as well
    :param size_threshold: every weight above that threshold is saved as external
    :param all_tensors_to_one_file: saves all tensors in one unique file
    """
    if not external:
        onnx.save_model(proto, filename)
        return

    dirname, shortname = os.path.split(filename)
    onnx.convert_model_to_external_data(
        proto,
        all_tensors_to_one_file=all_tensors_to_one_file,
        location=shortname + ".data",
        convert_attribute=convert_attribute,
        size_threshold=size_threshold,
    )

    proto = onnx.write_external_data_tensors(proto, dirname)
    with open(filename, "wb") as f:
        f.write(proto.SerializeToString())


def _info_type(
    typ: Union[onnx.TensorProto, onnx.TypeProto, onnx.SparseTensorProto]
) -> Dict[str, str]:
    if typ is None:
        return {}
    if isinstance(typ, (onnx.TensorProto, onnx.SparseTensorProto)):
        shape = [str(i) for i in typ.dims]
        return dict(
            type="tensor", elem_type=_rev_type[typ.data_type], shape="x".join(shape)
        )
    if typ.tensor_type:
        ret = dict(type="tensor", elem_type=_rev_type[typ.tensor_type.elem_type])
        shape = []
        for d in typ.tensor_type.shape.dim:
            if d.dim_value:
                shape.append(str(d.dim_value))
            else:
                shape.append(d.dim_param or "?")
        ret["shape"] = "x".join(shape)
        return ret

    return dict(kind=str(type(typ)))


def enumerate_onnx_node_types(
    model: Union[str, onnx.ModelProto, onnx.GraphProto],
    level: int = 0,
    shapes: Optional[Dict[str, onnx.TypeProto]] = None,
    external: bool = True,
) -> Generator[Dict[str, Union[str, float]], None, None]:
    """
    Looks into types for every node in a model.

    :param model: a string or a proto
    :param level: level (recursivity level)
    :param shapes: known shapes,
        returned by :func:onnx.shape_inference.infer_shapes`
    :param external: loads the external data if the model is loaded
    :return: a list of dictionary which can be turned into a dataframe.
    """
    proto = load_model(model, external=external)
    if shapes is None and isinstance(proto, onnx.ModelProto):
        p2 = onnx.shape_inference.infer_shapes(proto)
        values = p2.graph.value_info
        shapes = {}
        for value in values:
            shapes[value.name] = value.type
        for o in proto.graph.output:
            if o.name not in shapes:
                shapes[o.name] = o.type

    if isinstance(proto, onnx.ModelProto):
        if shapes is None:
            raise RuntimeError("shape inference has failed.")
        for item in enumerate_onnx_node_types(proto.graph, level=level, shapes=shapes):
            yield item

    elif isinstance(model, onnx.FunctionProto):
        raise NotImplementedError(f"Not implemented for type {type(proto)}.")

    else:
        for inp in proto.input:
            obs = dict(level=level, name=inp.name, kind="input")
            obs.update(_info_type(inp.type))
            yield obs

        for init in proto.initializer:
            obs = dict(level=level, name=init.name, kind="initializer")
            obs.update(_info_type(init))
            yield obs

        for init in proto.sparse_initializer:
            obs = dict(level=level, name=init.name, kind="sparse_initializer")
            obs.update(_info_type(init))
            yield obs

        for node in proto.node:
            obs = dict(
                level=level,
                name=node.name,
                kind="Op",
                domain=node.domain,
                type=node.op_type,
                inputs=",".join(node.input),
                outputs=",".join(node.output),
                input_types=",".join(
                    _info_type(shapes.get(i, None)).get("elem_type", "")
                    for i in node.input
                ),
                output_types=",".join(
                    _info_type(shapes.get(i, None)).get("elem_type", "")
                    for i in node.output
                ),
            )
            yield obs

            for att in node.attribute:
                if att.type == onnx.AttributeProto.GRAPH:
                    obs = dict(name=att.name, kind="attribute", level=level + 1)
                    yield obs
                    for item in enumerate_onnx_node_types(
                        att.g, level=level + 1, shapes=shapes
                    ):
                        yield item

            for out in node.output:
                obs = dict(name=out, kind="result", level=level)
                obs.update(_info_type(shapes.get(out, None)))
                yield obs

        for out in proto.output:
            obs = dict(level=level, name=out.name, kind="output")
            obs.update(_info_type(out.type))
            yield obs



import os
import sys
import unittest
import warnings
from argparse import ArgumentParser, Namespace
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO
from timeit import Timer
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy
from numpy.testing import assert_allclose


def unit_test_going():
    """
    Enables a flag telling the script is running while testing it.
    Avois unit tests to be very long.
    """
    going = int(os.environ.get("UNITTEST_GOING", 0))
    return going == 1


def ignore_warnings(warns: List[Warning]) -> Callable:
    """
    Catches warnings.

    :param warns:   warnings to ignore
    """

    def wrapper(fct):
        if warns is None:
            raise AssertionError(f"warns cannot be None for '{fct}'.")

        def call_f(self):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", warns)
                return fct(self)

        return call_f

    return wrapper


def measure_time(
    stmt: Union[str, Callable],
    context: Optional[Dict[str, Any]] = None,
    repeat: int = 10,
    number: int = 50,
    warmup: int = 1,
    div_by_number: bool = True,
    max_time: Optional[float] = None,
) -> Dict[str, Any]:
    """
    Measures a statement and returns the results as a dictionary.

    :param stmt: string or callable
    :param context: variable to know in a dictionary
    :param repeat: average over *repeat* experiment
    :param number: number of executions in one row
    :param warmup: number of iteration to do before starting the
        real measurement
    :param div_by_number: divide by the number of executions
    :param max_time: execute the statement until the total goes
        beyond this time (approximatively), *repeat* is ignored,
        *div_by_number* must be set to True
    :return: dictionary

    .. runpython::
        :showcode:

        from onnx_extended.ext_test_case import measure_time
        from math import cos

        res = measure_time(lambda: cos(0.5))
        print(res)

    See `Timer.repeat <https://docs.python.org/3/library/
    timeit.html?timeit.Timer.repeat>`_
    for a better understanding of parameter *repeat* and *number*.
    The function returns a duration corresponding to
    *number* times the execution of the main statement.

    .. versionchanged:: 0.4
        Parameter *max_time* was added.
    """
    if not callable(stmt) and not isinstance(stmt, str):
        raise TypeError(
            f"stmt is not callable or a string but is of type {type(stmt)!r}."
        )
    if context is None:
        context = {}

    if isinstance(stmt, str):
        tim = Timer(stmt, globals=context)
    else:
        tim = Timer(stmt)

    if warmup > 0:
        warmup_time = tim.timeit(warmup)
    else:
        warmup_time = 0

    if max_time is not None:
        if not div_by_number:
            raise ValueError(
                "div_by_number must be set to True of max_time is defined."
            )
        i = 1
        total_time = 0.0
        results = []
        while True:
            for j in (1, 2):
                number = i * j
                time_taken = tim.timeit(number)
                results.append((number, time_taken))
                total_time += time_taken
                if total_time >= max_time:
                    break
            if total_time >= max_time:
                break
            ratio = (max_time - total_time) / total_time
            ratio = max(ratio, 1)
            i = int(i * ratio)

        res = numpy.array(results)
        tw = res[:, 0].sum()
        ttime = res[:, 1].sum()
        mean = ttime / tw
        ave = res[:, 1] / res[:, 0]
        dev = (((ave - mean) ** 2 * res[:, 0]).sum() / tw) ** 0.5
        mes = dict(
            average=mean,
            deviation=dev,
            min_exec=numpy.min(ave),
            max_exec=numpy.max(ave),
            repeat=1,
            number=tw,
            ttime=ttime,
        )
    else:
        res = numpy.array(tim.repeat(repeat=repeat, number=number))
        if div_by_number:
            res /= number

        mean = numpy.mean(res)
        dev = numpy.mean(res**2)
        dev = (dev - mean**2) ** 0.5
        mes = dict(
            average=mean,
            deviation=dev,
            min_exec=numpy.min(res),
            max_exec=numpy.max(res),
            repeat=repeat,
            number=number,
            ttime=res.sum(),
        )

    if "values" in context:
        if hasattr(context["values"], "shape"):
            mes["size"] = context["values"].shape[0]
        else:
            mes["size"] = len(context["values"])
    else:
        mes["context_size"] = sys.getsizeof(context)
    mes["warmup_time"] = warmup_time
    return mes


class ExtTestCase(unittest.TestCase):
    _warns: List[Tuple[str, int, Warning]] = []

    def assertExists(self, name):
        if not os.path.exists(name):
            raise AssertionError(f"File or folder {name!r} does not exists.")

    def assertEqualArray(
        self,
        expected: numpy.ndarray,
        value: numpy.ndarray,
        atol: float = 0,
        rtol: float = 0,
    ):
        self.assertEqual(expected.dtype, value.dtype)
        self.assertEqual(expected.shape, value.shape)
        assert_allclose(expected, value, atol=atol, rtol=rtol)

    def assertAlmostEqual(
        self,
        expected: numpy.ndarray,
        value: numpy.ndarray,
        atol: float = 0,
        rtol: float = 0,
    ):
        if not isinstance(expected, numpy.ndarray):
            expected = numpy.array(expected)
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value).astype(expected.dtype)
        self.assertEqualArray(expected, value, atol=atol, rtol=rtol)

    def assertRaise(self, fct: Callable, exc_type: type[Exception]):
        try:
            fct()
        except exc_type as e:
            if not isinstance(e, exc_type):
                raise AssertionError(f"Unexpected exception {type(e)!r}.")
            return
        raise AssertionError("No exception was raised.")

    def assertEmpty(self, value: Any):
        if value is None:
            return
        if len(value) == 0:
            return
        raise AssertionError(f"value is not empty: {value!r}.")

    def assertNotEmpty(self, value: Any):
        if value is None:
            raise AssertionError(f"value is empty: {value!r}.")
        if isinstance(value, (list, dict, tuple, set)):
            if len(value) == 0:
                raise AssertionError(f"value is empty: {value!r}.")

    def assertStartsWith(self, prefix: str, full: str):
        if not full.startswith(prefix):
            raise AssertionError(f"prefix={prefix!r} does not start string  {full!r}.")

    @classmethod
    def tearDownClass(cls):
        for name, line, w in cls._warns:
            warnings.warn(f"\n{name}:{line}: {type(w)}\n  {str(w)}")

    def capture(self, fct: Callable):
        """
        Runs a function and capture standard output and error.

        :param fct: function to run
        :return: result of *fct*, output, error
        """
        sout = StringIO()
        serr = StringIO()
        with redirect_stdout(sout):
            with redirect_stderr(serr):
                res = fct()
        return res, sout.getvalue(), serr.getvalue()


def get_parsed_args(
    name: str,
    scenarios: Optional[Dict[str, str]] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    number: int = 10,
    repeat: int = 10,
    warmup: int = 5,
    sleep: float = 0.1,
    tries: int = 2,
    expose: Optional[str] = None,
    **kwargs: Dict[str, Tuple[Union[int, str, float], str]],
) -> Namespace:
    """
    Returns parsed arguments for examples in this package.

    :param name: script name
    :param scenarios: list of available scenarios
    :param description: parser description
    :param epilog: text at the end of the parser
    :param number: default value for number parameter
    :param repeat: default value for repeat parameter
    :param warmup: default value for warmup parameter
    :param sleep: default value for sleep parameter
    :param expose: if empty, keeps all the parameters,
        if None, only publish kwargs contains, otherwise the list
        of parameters to publish separated by a comma
    :param kwargs: additional parameters,
        example: `n_trees=(10, "number of trees to train")`
    :return: parser
    """
    if description is None:
        description = f"Available options for {name}.py."
    if epilog is None:
        epilog = ""
    parser = ArgumentParser(prog=name, description=description, epilog=epilog)
    if expose is not None:
        to_publish = set(expose.split(",")) if len(expose) > 0 else set()
        if scenarios is not None:
            rows = ", ".join(f"{k}: {v}" for k, v in scenarios.items())
            parser.add_argument(
                "-s", "--scenario", help=f"Available scenarios: {rows}."
            )
        if len(to_publish) == 0 or "number" in to_publish:
            parser.add_argument(
                "-n",
                "--number",
                help=f"number of executions to measure, default is {number}",
                type=int,
                default=number,
            )
        if len(to_publish) == 0 or "repeat" in to_publish:
            parser.add_argument(
                "-r",
                "--repeat",
                help=f"number of times to repeat the measure, default is {repeat}",
                type=int,
                default=repeat,
            )
        if len(to_publish) == 0 or "warmup" in to_publish:
            parser.add_argument(
                "-w",
                "--warmup",
                help=f"number of times to repeat the measure, default is {warmup}",
                type=int,
                default=warmup,
            )
        if len(to_publish) == 0 or "sleep" in to_publish:
            parser.add_argument(
                "-S",
                "--sleep",
                help=f"sleeping time between two configurations, default is {sleep}",
                type=float,
                default=sleep,
            )
        if len(to_publish) == 0 or "tries" in to_publish:
            parser.add_argument(
                "-t",
                "--tries",
                help=f"number of tries for each configurations, default is {tries}",
                type=int,
                default=tries,
            )
    for k, v in kwargs.items():
        parser.add_argument(
            f"--{k}",
            help=f"{v[1]}, default is {v[0]}",
            type=type(v[0]),
            default=v[0],
        )

    return parser.parse_args()
